# -*- coding: utf-8 -*-
"""
Created on Tue Dec 14 14:12:08 2021

@author: huzon
"""

import numpy as np
from tensorflow import keras
from tensorflow.keras import backend as K


def cosine_decay_with_warmup(global_step,
                             learning_rate_base,
                             total_steps,
                             warmup_learning_rate=0.0,
                             warmup_steps=0,
                             hold_base_rate_steps=0):
    """
    Parameters
    ----------
    global_step : TYPE
        DESCRIPTION.
    learning_rate_base : TYPE
        DESCRIPTION.
    total_steps : TYPE
        DESCRIPTION.
    warmup_learning_rate : TYPE, optional
        DESCRIPTION. The default is 0.0.
    warmup_steps : TYPE, optional
        DESCRIPTION. The default is 0.
    hold_base_rate_steps : TYPE, optional
        DESCRIPTION. The default is 0.

    Raises
    ------
    ValueError
        DESCRIPTION.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    
    if total_steps < warmup_steps:
         raise ValueError('total_steps must be larger or equal to '
                          'warmup_steps.')

    learning_rate = 0.5 * learning_rate_base * (1 + np.cos(np.pi *
        (global_step - warmup_steps - hold_base_rate_steps) / float(total_steps - warmup_steps - hold_base_rate_steps))) + 1e-7

    if hold_base_rate_steps > 0:
        learning_rate = np.where(global_step > warmup_steps + hold_base_rate_steps,
                                 learning_rate, learning_rate_base)
    if warmup_steps > 0:
        if learning_rate_base < warmup_learning_rate:
            raise ValueError('learning_rate_base must be larger or equal to '
                             'warmup_learning_rate.')

        slope = (learning_rate_base - warmup_learning_rate) / warmup_steps
        warmup_rate = slope * global_step + warmup_learning_rate
        
        learning_rate = np.where(global_step < warmup_steps, warmup_rate,
                                 learning_rate)
    return np.where(global_step > total_steps, 0.0, learning_rate)


class WarmUpCosineDecayScheduler(keras.callbacks.Callback):
    def __init__(self,
                 epochs=0,
                 restart_epoches=0,
                 train_per_epoch=0,
                 learning_rate_base=0,
                 global_step_init=0,
                 warmup_learning_rate=0.0,
                 warmup_steps=0,
                 hold_base_rate_steps=0,
                 verbose=0):
        super().__init__()
        self.restart_epoches = restart_epoches
        self.train_per_epoch = train_per_epoch
        self.learning_rate_base = learning_rate_base
        self.global_step = global_step_init
        self.warmup_learning_rate = warmup_learning_rate
        self.warmup_steps = warmup_steps
        self.hold_base_rate_steps = hold_base_rate_steps
        self.verbose = verbose
        self.learning_rates = []
        self.former_epochs = 0
        self.steps = int(epochs * train_per_epoch)
        if restart_epoches:
            self.epochs_hold = restart_epoches
            self.steps = int(restart_epoches * train_per_epoch)
            self.cumsum = restart_epoches
            self.cumul = restart_epoches
    
    
    def on_epoch_begin(self, epoch, logs=None):
        if self.restart_epoches:
            if epoch >= self.restart_epoches:
                self.global_step = 0
                self.former_epochs = self.restart_epoches
                self.cumsum = self.cumsum + self.cumul
                self.restart_epoches = self.restart_epoches + self.cumsum
                self.epochs_hold = self.restart_epoches - self.former_epochs
                self.steps = int(self.epochs_hold * self.train_per_epoch)


    def on_batch_end(self, batch, logs=None):
        """
        Parameters
        ----------
        batch : TYPE
            DESCRIPTION.
        logs : TYPE, optional
            DESCRIPTION. The default is None.

        Returns
        -------
        None.

        """
        self.global_step = self.global_step + 1
        lr = K.get_value(self.model.optimizer.lr)
        self.learning_rates.append(lr)

    
    def on_batch_begin(self, batch, logs=None):
        """
        Parameters
        ----------
        batch : TYPE
            DESCRIPTION.
        logs : TYPE, optional
            DESCRIPTION. The default is None.

        Returns
        -------
        None.

        """
        lr = cosine_decay_with_warmup(global_step=self.global_step,
                                      learning_rate_base=self.learning_rate_base,
                                      total_steps=self.steps,
                                      warmup_learning_rate=self.warmup_learning_rate,
                                      warmup_steps=self.warmup_steps,
                                      hold_base_rate_steps=self.hold_base_rate_steps)
        
        K.set_value(self.model.optimizer.lr, lr)
        if self.verbose > 0:
            print('\nBatch %05d: setting learning '
                  'rate to %s.' % (self.global_step + 1, lr))



# -*- coding: utf-8 -*-
"""
Created on Sat Oct  9 15:17:09 2021

@author: hzx
"""

import json
from tqdm import trange
import numpy as np
from typing import List, Dict, Any, Union
from mendeleev import element
from atom2vec import AtomSimilarity
from pymatgen.core import Structure, Element
from pymatgen.core.periodic_table import get_el_sp


class Atom2Vector:
    """
    Using principle of word2vec to embedding atom number to a vector.
    Default using e_above_hull to select structures in mp database.
    """
    def __init__(self, embedding_dim=16,
                 max_elements=8,
                 structures:List[Structure] = None):
        """
        Parameters
        ----------
        embedding_dim : TYPE, optional
            DESCRIPTION. The default is 20. Dimension of sematic vector of element.
        max_elements : TYPE, optional
            DESCRIPTION. The default is 10. The maximum number of elements in a structure.
        Returns
        -------
        None.

        """
        self.structures: List[Structure]
        self.structures = structures
        self.atom_similarity = self.cal_atom_similarity(self.structures, 
                                                     k_dim=embedding_dim, max_elements=max_elements)         


    @staticmethod 
    def cal_atom_similarity(structures:List, k_dim, max_elements):
        """
        Parameters
        ----------
        structures : List[Structure]
            DESCRIPTION. a list of pymatgen.core.Structure.
            You can transform poscar or cif file to pymatgen.core.Structure
            by using Structure.from_file(poscar or cif file).
        k_dim : TYPE
            DESCRIPTION.
        max_elements : TYPE
            DESCRIPTION.
        mpr : TYPE, optional
            DESCRIPTION. The default is True.

        Returns
        -------
        atom_similarity : TYPE
            DESCRIPTION.

        """
        atom_similarity = AtomSimilarity.from_structures(structures, k_dim, max_elements)
        return atom_similarity


    def get_atom_feature(self, element: Element) -> List[float]:
        """
        Parameters
        ----------
        element : pymatgen.core.Element
            DESCRIPTION.

        Returns
        -------
        List
            transform atom to a vector by atom2vector.
        """

        if isinstance(element, Element):
            atom_vector = self.atom_similarity.get_atom_vector(element)
            return atom_vector
        
        raise ValueError("%s not pymatgen.core.Element" % str(element))
        
        
    def get_atom_feature_table(self) -> Dict:
        """
        Returns
        -------
        Dict
            obtain all atom feature of periodic_table.
        """
        atom_vector_table = {}
        for Z in range(1,119):
            element = get_el_sp(Z)
            try:
                atom_vector_table[Z] = self.get_atom_feature(element)
            except:
                continue
        return atom_vector_table      


class Mendeleev_property:
    def __init__(self):
        pass
    
    @staticmethod
    def get_mendeleev_properties():
        """
        Returns
        -------
        dict
            DESCRIPTION.
            Get properties of elements in mendeleev table.
        """
        dicts = {'s':0, 'p':1, 'd':2, 'f':3}
        
        nvalence = {}
        unpaired_electrons = {}
        electronegativity = {}
        angle_quantum_num = {}
        ionenergies = {}
        dipole_polarizability = {}
        covalent_radius = {}
        atomic_weight = {}
        
        for i in trange(1,119):
            key = str(i)
            nvalence[key] = element(i).nvalence()
            unpaired_electrons[key] = element(i).ec.unpaired_electrons()
            if i <= 116:
                electronegativity[key] = element(i).electronegativity('pauling') or element(i-1).electronegativity('pauling') 
            else:
                electronegativity[key] = element(116).electronegativity('pauling')
            angle_quantum_num[key] = dicts[element(i).block]
            if i <= 110:
                ionenergies[key] = list(element(i).ionenergies.values())[0]
            else:
                ionenergies[key] = list(element(110).ionenergies.values())[0]
            dipole_polarizability[key] = element(i).dipole_polarizability
            covalent_radius[key] = element(i).covalent_radius
            atomic_weight[key] = element(i).atomic_weight
            
        return {"nvalence": nvalence,
                "unpaired_electrons": unpaired_electrons,
                "electronegativity": electronegativity, 
                "angle_quantum_num": angle_quantum_num,
                "ionenergies": ionenergies,
                "dipole_polarizability": dipole_polarizability,
                "covalent_radius":covalent_radius,
                "atomic_weight": atomic_weight}
    
    
    def save_mendeleev_properties(mendeleev_properties):
        with open('mendeleev.json','w') as f:
            json.dump(mendeleev_properties, f)


class Converter:
    
    def convert(d: Any) -> Any:
        """
        Parameters
        ----------
        d : Any
            DESCRIPTION.

        Returns
        -------
        None.

        """
        pass


class GaussianDistance(Converter):
    """
    Expand distance with Gaussian basis sit at centers and with width 0.5.
    """
    def __init__(self, centers: np.ndarray=np.linspace(0, 1, 64), width=0.5):
        """
        Parameters
        ----------
        centers : np.ndarray, optional
            DESCRIPTION. The default is np.linspace(0, 5, 100).
        width : TYPE, optional
            DESCRIPTION. The default is 0.5.

        Returns
        -------
        None.

        """
        self.centers = centers
        self.width = width


    def convert(self, distances: np.ndarray) -> np.ndarray:
        """
        Parameters
        ----------
        distances : np.ndarray
            expand distance vector distances with given parameters.

        Returns
        -------
        TYPE
            (matrix) N*M matrix with N the length of distances and M the length of centers.

        """
        distances = np.array(distances)
        return np.exp(-((distances[:, None] - self.centers[None, :])**2)/self.width**2)        


class MultiPropertyFeatures(Converter):
    """
    Get multi-property bond features.
    """
    def __init__(self, mendeleev_properties):
        self.properties = mendeleev_properties
        
    def convert(self, bond_indinces) -> np.ndarray:
        """
        Parameters
        ----------
        bond_indinces : TYPE
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.
            
        """
        bonds_properties = []
        for bond_indince in bond_indinces:
            start, end = str(bond_indince[0]), str(bond_indince[1])
            ven = self.properties["nvalence"][start]/self.properties["nvalence"][end] # valence eletron
            if start != end:
                upe = self.properties["unpaired_electrons"][start] + self.properties["unpaired_electrons"][end] # upair eletron
                aqn = self.properties["angle_quantum_num"][start] + self.properties["angle_quantum_num"][end] # angle quantum number
            else:
                upe = self.properties["unpaired_electrons"][start] # upair eletron
                aqn = self.properties["angle_quantum_num"][start] # angle quantum number
            eng = self.properties["electronegativity"][start]/self.properties["electronegativity"][end] # electronegativity
            fie = self.properties["ionenergies"][start]/self.properties["ionenergies"][end] # the first ionenergies
            plb = self.properties["dipole_polarizability"][start]/self.properties["dipole_polarizability"][end] # dipole_polarizability
            cvr = self.properties["covalent_radius"][start]/self.properties["covalent_radius"][end] # covalent_radius

            bonds_properties.append([ven, upe, aqn, eng, fie, plb, cvr])
        return np.array(bonds_properties, dtype=np.float32)
    
            
class Embedding_edges:
    """
    Costomed bond feature mbedding using different bond properties.
    """
    def __init__(self, converter: Union[str, Converter]):
        self.converter = converter
        
    def embedding(self, d: Any) -> Any:
        """
        Parameters
        ----------
        d : Any
            DESCRIPTION.

        Returns
        -------
        Any
            DESCRIPTION.

        """
        return self.converter.convert(d)
    
    


# -*- coding: utf-8 -*-
"""
Created on Sun Oct 10 15:33:33 2021

@author: hzx
"""

from typing import Sequence
import tensorflow as tf
from tensorflow.keras import layers
from tensorflow.keras import activations, initializers, regularizers, constraints
from crysnet.utils.spherical_harmonics import evaluate_spherical_harmonics


class SphericalBasisLayer(layers.Layer):
    
    def __init__(self, edge_dim=64,
                 num_spherical=7,
                 kernel_initializer="glorot_uniform",
                 kernel_regularizer=None,
                 kernel_constraint=None,
                 use_bias=False,
                 bias_initializer="zeros",
                 bias_regularizer=None,
                 bias_constraint=None,
                 **kwargs):
        super().__init__( **kwargs)
        self.edge_dim = edge_dim
        self.num_spherical = num_spherical
        self.kernel_initializer = kernel_initializer
        self.kernel_regularizer = kernel_regularizer
        self.kernel_constraint = kernel_constraint
        self.use_bias = use_bias
        self.bias_initializer = bias_initializer
        self.bias_regularizer = bias_regularizer
        self.bias_constraint = bias_constraint
    

    def build(self, input_shape):
        dense_dim = self.num_spherical**2

        with tf.name_scope("edge_aggregate"):
            self.kernel = self.add_weight(
                shape=(dense_dim, 
                    self.edge_dim),
                trainable=True,
                initializer=self.kernel_initializer,
                regularizer=self.kernel_regularizer,
                constraint=self.kernel_constraint,
                name='sph_kernel',
            )
            self.bias = self.add_weight(
                shape=(self.edge_dim,),
                trainable=True,
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                name='sph_bias',
            )
        
        self.built = True
        

    def spherical_coordiates(self, local_env, pair_indices):
        """
        Parameters
        ----------
        graph : Dict
            DESCRIPTION.

        Returns
        -------
        List
            return local spherical coordinate parameters theta and phi for every edges and their index.

        """
        local_env = tf.reshape(local_env, shape=(-1, 2, 3))
        sends, recives = pair_indices[:,0], pair_indices[:,1]
        edges_neighbor = tf.where(tf.expand_dims(sends,-1)==recives)
        polar = tf.gather(local_env, edges_neighbor[:,0])[:,0]
        edges = tf.gather(local_env, edges_neighbor[:,1])[:,0]
        vetical = tf.gather(local_env, edges_neighbor[:,0])[:,1]
        
        inner_product_theta = tf.reduce_sum(polar * edges, axis=-1)
        cross_product_theta = tf.linalg.cross(polar, edges)
        cross_product_theta = tf.norm(cross_product_theta, axis=-1)
        theta = tf.math.atan2(cross_product_theta, inner_product_theta)

        edges_projected = edges * tf.expand_dims(tf.cos(theta), axis=-1)  
        inner_product_phi = tf.reduce_sum(vetical * edges_projected, axis=-1)
        cross_product_phi = tf.linalg.cross(polar, edges_projected)
        cross_product_phi = tf.norm(cross_product_phi, axis=-1)
        phi = tf.math.atan2(cross_product_phi, inner_product_phi)
        
        phi_theta = tf.concat([tf.expand_dims(phi, -1), tf.expand_dims(theta, -1)], -1)
        
        return phi_theta, edges_neighbor


    def sph_harm_func(self, phi_theta):
        phi = phi_theta[:, 0]
        theta = phi_theta[:, 1]
        Ylm = tf.stack([
                evaluate_spherical_harmonics(l, m, theta, phi) for l in range(self.num_spherical) for m in range(-l, l + 1)])
        sph_harm_features = tf.transpose(Ylm, perm=(1, 0))
        return sph_harm_features

        
    def call(self, inputs):
        """
        Parameters
        ----------
        inputs : theta and phi of edges in local coordinates.

        Returns
        -------
        Ylm : spherical harmonics for local coordinates theta and phi.
        """
        local_env, pair_indices = inputs
        phi_theta, edges_neighbor = self.spherical_coordiates(local_env, pair_indices)
        # sph_harm_features = tf.cast(self.sph_harm_func(1,0,phi_theta[:,0],phi_theta[:,1]).real, dtype=tf.float32)
        sph_harm_features = self.sph_harm_func(phi_theta)
        sph_harm_features = tf.matmul(sph_harm_features, self.kernel) + self.bias
        return sph_harm_features, edges_neighbor


    def get_config(self):
        config = super().get_config()
        config.update({"units": self.units})
        config.update({"num_spherical", self.num_spherical})
        return config


class AzimuthLayer(layers.Layer):
    
    def __init__(self, units=16,
                 kernel_initializer="glorot_uniform",
                 kernel_regularizer=None,
                 kernel_constraint=None,
                 use_bias=False,
                 bias_initializer="zeros",
                 bias_regularizer=None,
                 bias_constraint=None,
                 **kwargs):
        super().__init__( **kwargs)
        self.units = units
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)

        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)

        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)
    
    
    def build(self, input_shape):
        if self.units is not None:
            dense_dim = self.units
        
        with tf.name_scope("edge_aggregate"):
            self.kernel = self.add_weight(
                shape=(2, dense_dim),
                trainable=True,
                initializer=self.kernel_initializer,
                regularizer=self.kernel_regularizer,
                constraint=self.kernel_constraint,
                name='azimuth_kernel',
            )

            self.bias = self.add_weight(
                shape=(dense_dim,),
                trainable=True,
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                name='azimuth_bias',
            )
        
        self.built = True
        

    def spherical_coordiates(self, local_env, pair_indices):
        """
        Parameters
        ----------
        graph : Dict
            DESCRIPTION.

        Returns
        -------
        List
            return local spherical coordinate parameters theta and phi for every edges and their index.

        """
        local_env = tf.reshape(local_env, shape=(-1, 2, 3))
        sends, recives = pair_indices[:,0], pair_indices[:,1]
        edges_neighbor = tf.where(tf.expand_dims(sends, -1)==recives)
        polar = tf.gather(local_env, edges_neighbor[:,0])[:,0]
        edges = tf.gather(local_env, edges_neighbor[:,1])[:,0]
        vetical = tf.gather(local_env, edges_neighbor[:,0])[:,1]
        
        inner_product_theta = tf.reduce_sum(polar * edges, axis=-1)
        cross_product_theta = tf.linalg.cross(polar, edges)
        cross_product_theta = tf.norm(cross_product_theta, axis=-1)
        theta = tf.math.atan2(cross_product_theta, inner_product_theta)

        edges_projected = edges*tf.expand_dims(tf.cos(theta), axis=-1)  
        inner_product_phi = tf.reduce_sum(vetical*edges_projected, axis=-1)
        cross_product_phi = tf.linalg.cross(polar, edges_projected)
        cross_product_phi = tf.norm(cross_product_phi, axis=-1)
        phi = tf.math.atan2(cross_product_phi, inner_product_phi)
        
        phi_theta = tf.concat([tf.expand_dims(phi, -1), tf.expand_dims(theta, -1)], -1)

        return phi_theta, edges_neighbor

        
    def call(self, inputs):
        """
        Parameters
        ----------
        inputs : theta and phi of edges in local coordinates.

        Returns
        -------
        Ylm : spherical harmonics for local coordinates theta and phi.
        """
        local_env, pair_indices = inputs
        phi_theta, edges_neighbor = self.spherical_coordiates(local_env, pair_indices)

        azimuth_features = tf.matmul(phi_theta, self.kernel) + self.bias
        return azimuth_features, edges_neighbor
    
    
    def get_config(self):
        config = super().get_config()
        config.update({"units": self.units})
        return config


class EdgeAggragate(layers.Layer):
    """
    This block is different from MessagePassomg block for aggregating and updating of Nodes,
    it is used to certain tasks that needs aggregating and updating edges of graph.
    """
    def __init__(self,
        units=64,
        kernel_initializer="glorot_uniform",
        kernel_regularizer=None,
        kernel_constraint=None,
        bias_initializer="zeros",
        bias_regularizer=None,
        bias_constraint=None,
        activation="relu",
        **kwargs):
        super().__init__(**kwargs)
        self.units = units
        self.activation = activations.get(activation)
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)

        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)

        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)
        

    def build(self, input_shape):
        self.edge_dim = input_shape[0][-1]
        
        with tf.name_scope("edge_aggregate_sph"):
            self.kernel = self.add_weight(
                shape=(self.edge_dim, self.units),
                trainable=True,
                initializer=self.kernel_initializer,
                regularizer=self.kernel_regularizer,
                constraint=self.kernel_constraint,
                name='edge_sph_kernel',
            )

            self.bias = self.add_weight(
                shape=(self.units),
                trainable=True,
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                name='edge_sph_bias',
            )
        
        self.built = True
    

    def call(self, inputs: Sequence) -> Sequence:
        """
        Parameters
        ----------
        inputs : Sequence
            DESCRIPTION.

        Returns
        -------
        Sequence
            DESCRIPTION.

        """
        edges_sph_features, edges_neighbor = inputs
        
        # concat scalar bond_features with directional edges_sph_features 
        edges_features_neighbors = tf.gather(edges_sph_features, edges_neighbor[:,1])
        # edges_feature_concated = tf.concat([edges_features_gather, edges_sph_features], axis=-1)
        
        # aggregate edges_features
        edges_features_aggregated = tf.math.segment_sum(edges_features_neighbors, edges_neighbor[:,0])
        transformed_edges_features = tf.matmul(edges_features_aggregated, self.kernel) + self.bias

        # return edges_features_updated
        return transformed_edges_features 


    def get_config(self):
        config = super().get_config()
        config.update({"units": self.units})
        return config


class ConcatLayer(layers.Layer):
    def __init__(self, units=16,
                kernel_initializer="glorot_uniform",
                kernel_regularizer=None,
                kernel_constraint=None,
                use_bias=True,
                bias_initializer="zeros",
                bias_regularizer=None,
                bias_constraint=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.units = units
        self.kernel_initializer = kernel_initializer
        self.kernel_regularizer = kernel_regularizer
        self.kernel_constraint = kernel_constraint
        self.use_bias = use_bias
        self.bias_initializer = bias_initializer
        self.bias_regularizer = bias_regularizer
        self.bias_constraint = bias_constraint


    def build(self, input_shape):
        self.bond_dim = input_shape[0][-1]
        self.sph_dim = input_shape[1][-1]

        with tf.name_scope("edge_concat"):
            self.kernel = self.add_weight(
                shape=(self.bond_dim + self.sph_dim, self.units),
                trainable=True,
                initializer=self.kernel_initializer,
                regularizer=self.kernel_regularizer,
                constraint=self.kernel_constraint,
                name='concat_kernel',
            )

            self.bias = self.add_weight(
                shape=(self.units),
                trainable=True,
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                name='concat_bias',
            )
        
        self.built = True


    def call(self, inputs):
        bond_features, edges_sph_features, edges_neighbor = inputs

        bond_features_gather = tf.gather(bond_features, edges_neighbor[:,1])
        edges_feature_concated = tf.concat([bond_features_gather, edges_sph_features], axis=-1)
        edges_features_aggregated = tf.math.segment_sum(edges_feature_concated, edges_neighbor[:,0])
        edges_features_aggregated = tf.matmul(edges_features_aggregated, self.kernel) + self.bias

        return edges_features_aggregated    


    def get_config(self):
        config = super().get_config()
        return config


class EdgeMessagePassing(layers.Layer):
    def __init__(self, units=64,
                 steps=3,
                 kernel_initializer="glorot_uniform",
                 kernel_regularizer=None,
                 kernel_constraint=None,
                 bias_initializer="zeros",
                 bias_regularizer=None,
                 bias_constraint=None,
                 activation="relu",
                 sph=False,
                 **kwargs):
        super().__init__(**kwargs)
        self.units = units
        self.steps = steps
        self.sph = sph


    def build(self, input_shape):
        self.edge_dim = input_shape[0][-1]
        self.azimuth_step = AzimuthLayer()
        self.sph_step = SphericalBasisLayer()
        self.concat = ConcatLayer(units=self.edge_dim)
        self.message_step = EdgeAggragate(self.units)
        self.update_step = layers.GRUCell(self.edge_dim, name='update_sph_edge')
        self.layernorm = layers.LayerNormalization()
        self.built = True


    def call(self, inputs):
        bond_features, local_env, pair_indices = inputs

        if self.sph:
            edges_sph_features, edges_neighbor = self.sph_step([local_env, pair_indices])
        else:
            edges_sph_features, edges_neighbor = self.azimuth_step([local_env, pair_indices])

        edges_features_updated = self.concat([bond_features, edges_sph_features, edges_neighbor])
        # edges_features_updated = bond_features
        
        # Perform a number of steps of message passing
        # Aggregate atom_features from neighbors
        for i in range(self.steps):
            edges_features_aggregated = self.message_step(
                [edges_features_updated, edges_neighbor]
            )
            
            # Update aggregated atom_features via a step of GRU
            edges_features_updated, _ = self.update_step(
                edges_features_aggregated, edges_features_updated
            )
                
        # edge_features_updated = self.layernorm(edge_features_updated)
        
        return edges_features_updated
    

    def get_config(self):
        config = super().get_config()
        config.update({"units": self.units})
        return config
    
    
class EdgeGraphNetwork(layers.Layer):
    def __init__(self,
        steps,
        kernel_initializer="glorot_uniform",
        kernel_regularizer=None,
        kernel_constraint=None,
        bias_initializer="zeros",
        bias_regularizer=None,
        bias_constraint=None,
        activation=None,
        sph=False,
        **kwargs):
        super().__init__(**kwargs)
        self.steps = steps
        self.activation = activations.get(activation)
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)

        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)

        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)

        self.sph = sph


    def build(self, input_shape):
        self.atom_dim = input_shape[0][-1]
        self.edge_dim = input_shape[1][-1]
        self.state_dim = input_shape[2][-1]

        self.azimuth_step = AzimuthLayer()
        self.sph_step = SphericalBasisLayer(edge_dim=self.edge_dim)
        self.concat = ConcatLayer()
        self.update_nodes = layers.GRUCell(self.atom_dim, name='update_nodes')
        self.update_edges = layers.GRUCell(self.edge_dim, name='update_edges')
        self.update_states = layers.GRUCell(self.state_dim, name='update_states')
        
        with tf.name_scope("edges_aggregate"):
            # weight for updating atom_features by bond_features 
            self.kernel = self.add_weight(
                shape=(self.atom_dim, 
                       (self.edge_dim + self.state_dim)**2),
                trainable=True,
                initializer=self.kernel_initializer,
                regularizer=self.kernel_regularizer,
                constraint=self.kernel_constraint,
                name='edges_kernel',
            )
            self.bias = self.add_weight(
                shape=((self.edge_dim + self.state_dim)**2,),
                trainable=True,
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                name='edges_bias',
            )
        
        self.built = True


    def concat_nodes(self, inputs: Sequence):
        """
        Parameters
        ----------
        inputs : TYPE
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        atom_features, edges_sph_features, state_attrs, pair_indices, atom_graph_indices, bond_graph_indices = inputs

        # each bond in pair_indices, concatenate thier atom features by atom indexes
        # then concatenate atom features with bond features
        edges_merge_sph_features = tf.math.segment_sum(edges_sph_features, pair_indices[:,0])
        
        # repeat state attributes by bond_graph_indices, then concatenate to bond_merge_atom_features
        state_attrs_repeat = tf.gather(state_attrs, atom_graph_indices)
        atoms_features_concated = tf.concat([atom_features, edges_merge_sph_features, state_attrs_repeat], axis=-1)

        return atoms_features_concated
        
    
    def aggregate_edges(self, inputs: Sequence):
        """
        Parameters
        ----------
        inputs : Sequence
            DESCRIPTION.

        Returns
        -------
        atom_features_aggregated : TYPE
            DESCRIPTION.

        """
        atom_features, edges_sph_features, state_attrs, edges_neighbor, atom_graph_indices, bond_graph_indices = inputs
        
        # concat state_attrs with edges_sph_features to get merged atom_merge_state_features
        state_attrs_repeat = tf.gather(state_attrs, bond_graph_indices)
        edge_merge_state_features = tf.concat([edges_sph_features, state_attrs_repeat], axis=-1)
        
        # using atom_updated to update edge_features_neighbors
        # using num_atoms of atom feature to renew num_atoms of adjacent edges feature
        # bond feature with shape (atom_dim,), not a Matrix, multiply by a learnable weight
        # with shape (edge_dim,edge_dim,atom_dim), then bond feature transfer to shape (edge_dim,edge_dim)
        # the bond matrix with shape (edge_dim,edge_dim) can update atom_feature with shape (edge_dim,)
        # so num_atoms of bond features need num_bonds of bond matrix, so a matrix with shape
        # (num_atoms,(edge_dim,edge_dim,atom_dim)) to transfer atom_features to shape (num_atoms,(edge_dim,edge_dim))
        # finally, apply this atom_matrix to adjacent edges, get atom_features updated edge_features_neighbors
        atom_features_gather = tf.gather(atom_features, tf.gather(pair_indices, edges_neighbor[:, 1])[:, 0])
        atom_weights = tf.matmul(atom_features_gather, self.kernel) + self.bias
        atom_weights = tf.reshape(atom_weights, (-1, self.edge_dim + self.state_dim, self.edge_dim + self.state_dim))
        edge_features_neighbors = tf.gather(edge_merge_state_features, edges_neighbor[:, 1])
        edge_features_neighbors = tf.expand_dims(edge_features_neighbors, axis=-1)
        transformed_features = tf.matmul(atom_weights, edge_features_neighbors)
        transformed_features = tf.squeeze(transformed_features, axis=-1)
        
        # using conbination of tf.operation realizes multiplicationf between adjacent matrix and atom features
        # first tf.gather end features using end atom index pair_indices[:,1] to atom_features_neighbors
        # then using bond matrix updates atom_features_neighbors, get transformed_features
        # finally tf.segment_sum calculates sum of updated neighbors feature by start atom index pair_indices[:,0]
        edge_features_aggregated = tf.math.segment_sum(transformed_features, edges_neighbor[:,0])

        return edge_features_aggregated
        
    
    def concat_states(self, inputs: Sequence):
        """
        Parameters
        ----------
        inputs : Sequence
            DESCRIPTION.

        Returns
        -------
        state_attrs_aggregated : TYPE
            DESCRIPTION.

        """
        atom_features, edges_sph_features, state_attrs, pair_indices, atom_graph_indices, bond_graph_indices = inputs



        # concat state_attrs with bond_updated and atom_features_aggregated
        edges_features_sum = tf.math.segment_sum(edges_sph_features, bond_graph_indices)
        atom_features_sum = tf.math.segment_sum(atom_features, atom_graph_indices)
        state_attrs_concated = tf.concat([atom_features_sum, edges_features_sum, state_attrs], axis=-1)

        return state_attrs_concated


    def call(self, inputs: Sequence) -> Sequence:
        """
        Parameters
        ----------
        inputs : Sequence
            DESCRIPTION.

        Returns
        -------
        Sequence
            DESCRIPTION.

        """
        atom_features, bond_features, local_env, state_attrs, pair_indices, atom_graph_indices, bond_graph_indices = inputs

        atom_features_updated = atom_features
        state_attrs_updated = state_attrs
        
        if self.sph:
            edges_sph_features, edges_neighbor = self.sph_step([local_env, pair_indices])
        else:
            edges_sph_features, edges_neighbor = self.azimuth_step([local_env, pair_indices])

        edges_features_updated = self.concat([bond_features, edges_sph_features, edges_neighbor])

        # Perform a number of steps of message passing
        for i in range(self.steps):
            # concate and update bond_features
            atoms_features_concated = self.concat_nodes(
                            [atom_features_updated, edges_features_updated,
                            state_attrs_updated, edges_neighbor, 
                            atom_graph_indices, bond_graph_indices]
                            )
            
            # Update edge_features via a step of GRU
            atoms_features_updated, _ = self.update_nodes(atoms_features_concated, atom_features_updated)
            
            # Aggregate atom_features from neighbors
            edges_features_aggregated = self.aggregate_edges(
                                    [atom_features_updated, edges_features_updated,
                                    state_attrs_updated, pair_indices,
                                    atom_graph_indices, bond_graph_indices]
                                    )

            # Update aggregated atom_features via a step of GRU
            edges_features_updated, _ = self.update_edges(edges_features_aggregated, edges_features_updated)

            # update state_attrs
            state_attrs_concated = self.concat_states(
                                [atom_features_updated, edges_features_updated,
                                 state_attrs_updated, pair_indices,
                                 atom_graph_indices, bond_graph_indices]
                                )

            state_attrs_updated, _ = self.update_states(state_attrs_concated, state_attrs_updated)
            
        return [atom_features_updated, edges_features_updated, state_attrs_updated]
    
    
    def get_config(self):
        config = super().get_config()
        config.update({"steps": self.steps})
        return config    
    


# -*- coding: utf-8 -*-
"""
Created on Sun Oct 10 15:33:33 2021

@author: hzx
"""

from typing import Sequence
import tensorflow as tf
from tensorflow.keras import layers
from tensorflow.keras import activations, initializers, regularizers, constraints


class MessagePassing(layers.Layer):
    def __init__(self,
        steps,
        kernel_initializer="glorot_uniform",
        kernel_regularizer=None,
        kernel_constraint=None,
        bias_initializer="zeros",
        bias_regularizer=None,
        recurrent_regularizer=None,
        bias_constraint=None,
        activation=None,
        **kwargs):
        super().__init__(**kwargs)
        self.steps = steps
        self.activation = activations.get(activation)
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)

        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.recurrent_regularizer = regularizers.get(recurrent_regularizer)

        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)


    def build(self, input_shape):
        self.atom_dim = input_shape[0][-1]
        self.edge_dim = input_shape[1][-1]
        self.state_dim = input_shape[2][-1]

        self.update_nodes = layers.GRUCell(self.atom_dim,
                                            kernel_regularizer=self.recurrent_regularizer,
                                            recurrent_regularizer=self.recurrent_regularizer,
                                            name='update_nodes'
                                            )

        self.update_edges = layers.GRUCell(self.edge_dim,
                                            kernel_regularizer=self.recurrent_regularizer,
                                            recurrent_regularizer=self.recurrent_regularizer,
                                            name='update_edges'
                                            )

        self.update_states = layers.GRUCell(self.state_dim,
                                            kernel_regularizer=self.recurrent_regularizer,
                                            recurrent_regularizer=self.recurrent_regularizer,
                                            name='update_states')
        
        with tf.name_scope("nodes_aggregate"):
            # weight for updating atom_features by bond_features 
            self.kernel = self.add_weight(
                shape=(self.edge_dim, 
                       (self.atom_dim + self.state_dim)**2),
                trainable=True,
                initializer=self.kernel_initializer,
                regularizer=self.kernel_regularizer,
                constraint=self.kernel_constraint,
                name='nodes_kernel',
            )
            self.bias = self.add_weight(
                shape=((self.atom_dim + self.state_dim)**2,),
                trainable=True,
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                name='nodes_bias',
            )
        
        self.built = True


    def concat_edges(self, inputs: Sequence):
        """
        Parameters
        ----------
        inputs : TYPE
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        atom_features, edges_sph_features, state_attrs, pair_indices, atom_graph_indices, bond_graph_indices = inputs

        # each bond in pair_indices, concatenate thier atom features by atom indexes
        # then concatenate atom features with bond features
        atom_features_gather = tf.gather(atom_features, pair_indices)
        edges_merge_atom_features = tf.concat([atom_features_gather[:,0], atom_features_gather[:,1]], axis=-1)
        
        # repeat state attributes by bond_graph_indices, then concatenate to bond_merge_atom_features
        state_attrs_repeat = tf.gather(state_attrs, bond_graph_indices)
        edges_features_concated = tf.concat([edges_merge_atom_features, state_attrs_repeat, edges_sph_features], axis=-1)

        return edges_features_concated
        
    
    def aggregate_nodes(self, inputs: Sequence):
        """
        Parameters
        ----------
        inputs : Sequence
            DESCRIPTION.

        Returns
        -------
        atom_features_aggregated : TYPE
            DESCRIPTION.

        """
        atom_features, edges_sph_features, state_attrs, pair_indices, atom_graph_indices, bond_graph_indices = inputs
        
        # concat state_attrs with atom_features to get merged atom_merge_state_features
        state_attrs_repeat = tf.gather(state_attrs, atom_graph_indices)
        atom_merge_state_features = tf.concat([atom_features, state_attrs_repeat], axis=-1)
        
        # using bond_updated to update atom_features_neighbors
        # using num_bonds of bond feature to renew num_bonds of adjacent atom feature
        # bond feature with shape (bond_dim,), not a Matrix, multiply by a learnable weight
        # with shape (atom_dim,atom_dim,bond_dim), then bond feature transfer to shape (atom_dim,atom_dim)
        # the bond matrix with shape (atom_dim,atom_dim) can update atom_feature with shape (aotm_dim,)
        # so num_bonds of bond features need num_bonds of bond matrix, so a matrix with shape
        # (num_bonds,(atom_dim,atom_dim,bond_dim)) to transfer bond_features to shape (num_bonds,(atom_dim,atom_dim))
        # finally, apply this bond_matrix to adjacent atoms, get bond_features updated atom_features_neighbors
        edges_weights = tf.matmul(edges_sph_features, self.kernel) + self.bias
        edges_weights = tf.reshape(edges_weights, (-1, self.atom_dim + self.state_dim, self.atom_dim + self.state_dim))
        atom_features_neighbors = tf.gather(atom_merge_state_features, pair_indices[:, 1])
        atom_features_neighbors = tf.expand_dims(atom_features_neighbors, axis=-1)
        transformed_features = tf.matmul(edges_weights, atom_features_neighbors)
        transformed_features = tf.squeeze(transformed_features, axis=-1)
        
        # using conbination of tf.operation realizes multiplicationf between adjacent matrix and atom features
        # first tf.gather end features using end atom index pair_indices[:,1] to atom_features_neighbors
        # then using bond matrix updates atom_features_neighbors, get transformed_features
        # finally tf.segment_sum calculates sum of updated neighbors feature by start atom index pair_indices[:,0]
        atom_features_aggregated = tf.math.segment_sum(transformed_features, pair_indices[:,0])

        return atom_features_aggregated
        
    
    def concat_states(self, inputs: Sequence):
        """
        Parameters
        ----------
        inputs : Sequence
            DESCRIPTION.

        Returns
        -------
        state_attrs_aggregated : TYPE
            DESCRIPTION.

        """
        atom_features, edges_sph_features, state_attrs, pair_indices, atom_graph_indices, bond_graph_indices = inputs

        # concat state_attrs with bond_updated and atom_features_aggregated
        edges_features_sum = tf.math.segment_sum(edges_sph_features, bond_graph_indices)
        atom_features_sum = tf.math.segment_sum(atom_features, atom_graph_indices)
        state_attrs_concated = tf.concat([atom_features_sum, edges_features_sum, state_attrs], axis=-1)

        return state_attrs_concated


    def call(self, inputs: Sequence) -> Sequence:
        """
        Parameters
        ----------
        inputs : Sequence
            DESCRIPTION.

        Returns
        -------
        Sequence
            DESCRIPTION.

        """
        atom_features, edges_sph_features, state_attrs, pair_indices, atom_graph_indices, bond_graph_indices = inputs
        
        atom_features_updated = atom_features
        edges_features_updated =  edges_sph_features
        state_attrs_updated = state_attrs

        # Perform a number of steps of message passing
        for i in range(self.steps):
            # concate and update bond_features
            edges_features_concated = self.concat_edges(
                            [atom_features_updated, edges_features_updated,
                            state_attrs_updated, pair_indices, 
                            atom_graph_indices, bond_graph_indices]
                            )
            
            # Update edge_features via a step of GRU
            edges_features_updated, _ = self.update_edges(edges_features_concated, edges_features_updated)
            
            # Aggregate atom_features from neighbors
            atom_features_aggregated = self.aggregate_nodes(
                                    [atom_features_updated, edges_features_updated,
                                    state_attrs_updated, pair_indices,
                                    atom_graph_indices, bond_graph_indices]
                                    )

            # Update aggregated atom_features via a step of GRU
            atom_features_updated, _ = self.update_nodes(atom_features_aggregated, atom_features_updated)

            # update state_attrs
            state_attrs_concated = self.concat_states(
                                [atom_features_updated, edges_features_updated,
                                 state_attrs_updated, pair_indices,
                                 atom_graph_indices, bond_graph_indices]
                                )

            state_attrs_updated, _ = self.update_states(state_attrs_concated, state_attrs_updated)
            
        return [atom_features_updated, edges_features_updated, state_attrs_updated]
    

    def get_config(self):
        config = super().get_config()
        config.update({"steps": self.steps})
        return config    
    
    


# -*- coding: utf-8 -*-
"""
Created on Wed Jun  9 11:39:28 2021

@author: hzx
"""

import time
import json
import numpy as np
from tqdm import tqdm
from operator import itemgetter
from multiprocessing import Pool
import tensorflow as tf
from tensorflow.keras.utils import Sequence
from crysnet.utils import Features
from typing import Union, Dict, List, Set
from .embedding import Mendeleev_property, GaussianDistance, MultiPropertyFeatures, Embedding_edges
from pymatgen.core import Structure
from pymatgen.analysis.local_env import NearNeighbors, VoronoiNN
from crysnet.utils.get_nn import get_nn_info
from crysnet.utils import get_space_group_number


class LabelledCrystalGraphBase():
    
    def __init__(self, strategy: Union[None, NearNeighbors]=VoronoiNN(cutoff=18.0)
                 ):
        """
        Parameters
        ----------
        strategy : Union[str, NearNeighbors], optional
            DESCRIPTION. The default is 'VoronoiNN'.
        atom2vector : Union[None, Atom2Vector], optional
            DESCRIPTION. The default is None. if atom2vector is None, program is going to download dataset
            from materilsproject.org to construct Semantic space of atoms, it will cost a lot of time, 
            so recommand set atom2vector=Atom2Vector() at first.

        Raises
        ------
        RuntimeError
            DESCRIPTION.

        Returns
        -------
        None.

        """
    
        if isinstance(strategy, NearNeighbors):
            self.strategy = strategy
        
        self.properties = None
        with open('mendeleev.json','r') as f:
            self.properties = json.load(f)
        if self.properties is None:
            self.properties = Mendeleev_property.get_mendeleev_properties()
        
        
    def get_graph(self, structure: Structure) -> Dict:
        """
        Parameters
        ----------
        structure : pymatgen.core.structure.Structure
            Feed with pymatgen.core.structure.Structure and produce the graph.

        Raises
        ------
        RuntimeError
            DESCRIPTION.

        Returns
        -------
        Dict
        Labelled graph
            state_attributes is global attributes, here is symmetry,
            oxide_states in order of atoms,
            atoms are nodes of graph,
            bond are distance between nodes,
            image are direction of bond along abc,
            pair_indices are nodes indices of edges, including self-cycle edge,
            lattice are abc of structure,
            cart_coords are Descartes coordinations.

        """
        lattice = np.array(structure.as_dict()['lattice']['matrix'])
        cart_coords = structure.cart_coords
        space_group_number = get_space_group_number(structure)
        state_attributes = np.array([space_group_number],dtype="int32")
        
        node1 = []
        node2 = []
        bonds = []
        images = []
        for atom, neighbors in enumerate(self.strategy.get_all_nn_info(structure)):
            node1.extend([atom] * (len(neighbors) + 1))
            node2.append(atom)
            bonds.append(0.0)
            images.append((0,0,0))
            for neighbor in neighbors:
                node2.append(neighbor["site_index"])
                bonds.append(neighbor["weight"])
                images.append(neighbor['image'])
        atoms = self.get_Z_number(structure)
        pair_indices = list(zip(node1,node2))
        if np.size(np.unique(node1)) < len(atoms):
            raise RuntimeError("Isolated atoms found in the structure")

        return {Features.atom: atoms, Features.bond: bonds, Features.state: state_attributes,
                Features.pair_indices: pair_indices, Features.image: images,
                Features.lattice: lattice, Features.cart_coords: cart_coords}

    
    @staticmethod
    def get_Z_number(structure: Structure) -> List:
        """
        Parameters
        ----------
        structure : Structure
            Get atomic number from pymatgen.core.struture.Structure.
            structure.atomic_number can also get it.
        Returns
        -------
        List
            DESCRIPTION.
        """
        return np.array([i.specie.Z for i in structure], dtype="int32").tolist()  
 

    def _local_coordinates(self, graph: Dict) -> List: 
        """
        Parameters
        ----------
        graph : Dict
            DESCRIPTION.

        Returns
        -------
        TYPE
            a seires of polar coordinates used for building local coordinate.

        """
        pair_indices = graph[Features.pair_indices]
        images = graph[Features.image]
        lattice = graph[Features.lattice]
        a, b, c = lattice[0], lattice[1], lattice[2]
        cart_coords = graph[Features.cart_coords]
        
        local_env = []
        for idx, pair_indice in enumerate(pair_indices):
            image = images[idx]
            node_send, node_recive = pair_indice[1], pair_indice[0]
            polar = cart_coords[node_recive] - cart_coords[node_send] - \
                    a*image[0] - b*image[1] - c*image[2]
            vetical = np.array([polar[1], -polar[0], 0.0])
            local_env.append(np.array([polar[0], polar[1], polar[2], vetical[0], vetical[1], vetical[2]]))
            
        return local_env

    
    def graph_to_input(self, graph: Dict) -> List[np.ndarray]:
        """
        Parameters
        ----------
        graph : Dict
            DESCRIPTION.

        Returns
        -------
        list
            DESCRIPTION.

        """
        atom_num_pairs = [[graph[Features.atom][pair[0]], graph[Features.atom][pair[1]]] for pair in graph[Features.pair_indices]]
        distance_features = Embedding_edges(converter=GaussianDistance()).embedding(graph[Features.bond])
        multi_properties = Embedding_edges(converter=MultiPropertyFeatures(self.properties)).embedding(atom_num_pairs)
        bond_features = np.concatenate([distance_features, multi_properties], axis=1)

        local_env = self._local_coordinates(graph)
        
        return [
            np.array(graph[Features.atom], dtype=np.int32),
            np.array(bond_features),
            np.array(graph[Features.state], dtype=np.int32),
            np.array(graph[Features.pair_indices], dtype=np.int32),
            np.array(local_env),
            ]


    def structure_to_input(self, structure: Structure) -> List:
        """
        Parameters
        ----------
        structure : pymatgen.core.structure.Structure
            DESCRIPTION.

        Returns
        -------
        List
            DESCRIPTION.

        """
        graph = self.get_graph(structure)
        return self.graph_to_input(graph)

    
    def ragged_inputs_from_strcutre_list(self, structure_list: List) -> Set:
        """
        Parameters
        ----------
        structure_list : List
            a list of pymatgen.core.Structure
            In order to keep with Semantic space of atom2vector, MPRester should be used
            to get structures from materialsproject.

        Returns
        -------
        Set
            DESCRIPTION.

        """
        # Initialize graphs
        atom_features_list = []
        bond_features_list = []
        state_attrs_list = []
        pair_indices_list = []
        local_env_list = []
        for structure in tqdm(structure_list, desc="Generating labelled multi-property graphs"):
            atom_features, bond_features, state_attrs, pair_indices, local_env = self.structure_to_input(structure)
            atom_features_list.append(atom_features)
            bond_features_list.append(bond_features)
            state_attrs_list.append(state_attrs)
            pair_indices_list.append(pair_indices)
            local_env_list.append(local_env)
  
        return (
            tf.ragged.constant(atom_features_list, dtype=tf.int32),
            tf.ragged.constant(bond_features_list, dtype=tf.float32),
            tf.ragged.constant(local_env_list, dtype=tf.float32),
            tf.ragged.constant(state_attrs_list, dtype=tf.int32),
            tf.ragged.constant(pair_indices_list, dtype=tf.int64),
            )


    def inputs_from_strcutre_list(self, structure_list: List) -> Set:
        """
        Parameters
        ----------
        structure_list : List
            a list of pymatgen.core.Structure
            In order to keep with Semantic space of atom2vector, MPRester should be used
            to get structures from materialsproject.

        Returns
        -------
        Set
            DESCRIPTION.

        """
        # Initialize graphs
        atom_features_list = []
        bond_features_list = []
        state_attrs_list = []
        pair_indices_list = []
        local_env_list = []
        for structure in tqdm(structure_list, desc="Generating labelled directionial spherical harmonic graphs"):
            atom_features, bond_features, state_attrs, pair_indices, local_env = self.structure_to_input(structure)
            atom_features_list.append(atom_features)
            bond_features_list.append(bond_features)
            state_attrs_list.append(state_attrs)
            pair_indices_list.append(pair_indices)
            local_env_list.append(local_env)
  
        return (
            atom_features_list,
            bond_features_list,
            local_env_list,
            state_attrs_list,
            pair_indices_list,
            )


    def graphs_from_strcutre_list(self, structure_list: List) -> List:
        """
        Parameters
        ----------
        structure_list : List
            a list of pymatgen.core.Structure
            In order to keep with Semantic space of atom2vector, MPRester should be used
            to get structures from materialsproject.

        Returns
        -------
        List
            DESCRIPTION.

        """
        # Initialize graphs
        graphs = []
        for structure in tqdm(structure_list, desc="Generating labelled directionial spherical harmonic graphs"):
            graphs.append(self.structure_to_input(structure))
  
        return graphs


class LabelledCrystalGraph(LabelledCrystalGraphBase):
    def __init__(self, cutoff=3.0):
        self.cutoff = cutoff
        

    def graph_to_input(self, graph: Dict) -> List[np.ndarray]:
        """
        Parameters
        ----------
        graph : Dict
            DESCRIPTION.

        Returns
        -------
        list
            DESCRIPTION.

        """
        distance_features = Embedding_edges(converter=GaussianDistance()).embedding(graph[Features.bond])
        local_env = self._local_coordinates(graph)
        
        return [
            np.array(graph[Features.atom], dtype=np.int32),
            np.array(distance_features),
            np.array(graph[Features.state], dtype=np.int32),
            np.array(graph[Features.pair_indices], dtype=np.int32),
            np.array(local_env),
            ]


    def get_graph(self, structure: Structure) -> Dict:
        """
        Parameters
        ----------
        structure : Structure
            DESCRIPTION.
        space_group_number : int
            DESCRIPTION.

        Raises
        ------
        RuntimeError
            DESCRIPTION.

        Returns
        -------
        dict
            DESCRIPTION.

        """
        lattice = np.array(structure.as_dict()['lattice']['matrix'], dtype=np.float32)
        cart_coords = structure.cart_coords.astype(np.float32)
        space_group_number = get_space_group_number(structure) - 1
        state_attributes = np.array([space_group_number], dtype="int32")
        center_indices, neighbor_indices, images, bonds = get_nn_info(structure, cutoff=self.cutoff)
        atoms = self.get_Z_number(structure)
        pair_indices = np.concatenate([center_indices, neighbor_indices], axis=0).reshape(2, -1).transpose()

        return {Features.atom: atoms, Features.bond: bonds, Features.state: state_attributes,
                Features.pair_indices: pair_indices, Features.image: images,
                Features.lattice: lattice, Features.cart_coords: cart_coords}

    
    def _local_coordinates(self, graph: Dict) -> np.ndarray: 
        """
        calculate local environment using numpy array only.
        Parameters
        ----------
        graph : Dict
            DESCRIPTION.

        Returns
        -------
        TYPE
            a seires of polar coordinates used for building local coordinate.

        """
        pair_indices = graph[Features.pair_indices]
        images = graph[Features.image]
        lattice = graph[Features.lattice]
        a, b, c = lattice[0], lattice[1], lattice[2]
        cart_coords = graph[Features.cart_coords]
        it1 = itemgetter(pair_indices[:,0])
        it2 = itemgetter(pair_indices[:,1])
        recive=it1(cart_coords)
        send=it2(cart_coords)
        polar = recive - send - np.expand_dims(images[:,0], axis=-1)*a - np.expand_dims(images[:,1], axis=-1)*b - np.expand_dims(images[:,2], axis=-1)*c
        zeros= np.ones_like(np.expand_dims(polar[:,1], axis=-1))
        vetical = np.concatenate([np.expand_dims(polar[:,1], axis=-1), -np.expand_dims(polar[:,0], axis=-1), zeros], axis=-1)
        local_env = np.concatenate([polar, vetical], axis=-1)
        return local_env


    def inputs_from_strcutre_list(self, structure_list: List) -> Set:
        """
        Parameters
        ----------
        structure_list : List
            a list of pymatgen.core.Structure
            In order to keep with Semantic space of atom2vector, MPRester should be used
            to get structures from materialsproject.

        Returns
        -------
        Set
            DESCRIPTION.

        """
        # Initialize graphs
        start = time.time()
        pool = Pool()
        graphs = pool.map(self.structure_to_input, structure_list)
        pool.close()
        pool.join()
        
        atom_features_list = []
        bond_features_list = []
        state_attrs_list = []
        pair_indices_list = []
        local_env_list = []
        for graph in graphs:
            atom_features, bond_features, state_attrs, pair_indices, local_env = graph
            atom_features_list.append(atom_features)
            bond_features_list.append(bond_features)
            state_attrs_list.append(state_attrs)
            pair_indices_list.append(pair_indices)
            local_env_list.append(local_env)
        
        end = time.time()
        run_time = end - start
        print('run time: ', run_time)
        
        return (
            atom_features_list,
            bond_features_list,
            local_env_list,
            state_attrs_list,
            pair_indices_list,
            )    
    
    
class  GraphBatchGeneratorSequence(Sequence):   
    def __init__(self, atom_features_list: List[np.ndarray],
                bond_features_list: List[np.ndarray],
                local_env_list: List[np.ndarray],
                state_attrs_list: List[np.ndarray],
                pair_indices_list: List[np.ndarray],
                labels: Union[List, None]=None,
                task_type=None,
                batch_size=32,
                is_shuffle=False):
        """
        Parameters
        ----------
        X_tensor : TYPE
            DESCRIPTION.
        y_label : TYPE
            DESCRIPTION.
        batch_size : TYPE, optional
            DESCRIPTION. The default is 64.
        is_shuffle : TYPE, optional
            DESCRIPTION. The default is False.

        Returns
        -------
        None.

        """
        self.task_type = task_type
        self.data_size = len(atom_features_list)
        self.batch_size = batch_size
        self.total_index = np.arange(self.data_size)

        self.atom_features_list = atom_features_list
        self.bond_features_list = bond_features_list
        self.local_env_list = local_env_list
        self.state_attrs_list = state_attrs_list
        self.pair_indices_list = pair_indices_list

        self.labels = labels

        if is_shuffle:
            shuffle = itemgetter(np.random.permutation(self.total_index))
            self.total_index = shuffle(self.total_index)
    
    
    def __len__(self) -> int:
        return int(np.ceil(self.data_size / self.batch_size))


    def on_epoch_end(self):
        """
        code to be executed on epoch end
        """
        self.total_index = np.random.permutation(self.total_index)


    def __getitem__(self, index: int) -> tuple:
        batch_index = self.total_index[index * self.batch_size : (index + 1) * self.batch_size]
        get = itemgetter(*batch_index)

        atom_features_list = get(self.atom_features_list)
        bond_features_list = get(self.bond_features_list)
        local_env_list =  get(self.local_env_list)
        state_attrs_list = get(self.state_attrs_list)
        pair_indices_list = get(self.pair_indices_list)

        inputs_batch = (atom_features_list,
                        bond_features_list,
                        local_env_list,
                        state_attrs_list,
                        pair_indices_list,
                        )

        x_batch = self._merge_batch(inputs_batch)
        if self.labels is None:
            return x_batch
        y_batch = np.array(get(self.labels))

        return x_batch, (y_batch)


    # def __getitem__(self, index: int) -> tuple:
    #     batch_index = self.total_index[index * self.batch_size : (index + 1) * self.batch_size]
    #     get = itemgetter(*batch_index)

    #     atom_features_list = list(get(self.atom_features_list))
    #     bond_features_list = list(get(self.bond_features_list))
    #     local_env_list =  list(get(self.local_env_list))
    #     state_attrs_list = list(get(self.state_attrs_list))
    #     pair_indices_list = list(get(self.pair_indices_list))

    #     inputs_batch = (tf.ragged.constant(atom_features_list, dtype=tf.int32),
    #         tf.ragged.constant(bond_features_list, dtype=tf.float32),
    #         tf.ragged.constant(local_env_list, dtype=tf.float32),
    #         tf.ragged.constant(state_attrs_list, dtype=tf.int32),
    #         tf.ragged.constant(pair_indices_list, dtype=tf.int64),
    #         )

    #     x_batch = self._merge_batch(inputs_batch)
    #     y_batch = np.atleast_2d(get(self.labels))

    #     return x_batch, y_batch


    def _merge_batch(self, x_batch: tuple) -> tuple:
        """
        Merging a batch of graphs into a disconnected graph should reindex atoms only
        features of graphs desn't be changed only merge them to one dimension of globl graph.
        reindex indices in pair_indices by adding increment of number of atoms in the batch
        atom marked with structure indice also need to be tell in globl graph.
        Parameters
        ----------
        x_batch : TYPE
            DESCRIPTION.

        Returns
        -------
        atom_features : TYPE
            DESCRIPTION.
        bond_features : TYPE
            DESCRIPTION.
        state_attributes : TYPE
            DESCRIPTION.
        pair_indices : TYPE
            DESCRIPTION.
        atom_partition_indices: TYPE
            DESCRIPTION.
        bond_partition_indices: TYPE
            DESCRIPTION.
        """
        atom_features, bond_features, local_env, state_attrs, pair_indices = x_batch
    
        # Obtain number of atoms and bonds for each graph
        # allocate graph (structure) indice for atom and bond in global graph
        num_atoms_per_graph = []
        atom_graph_indices = []
        for i, atoms in enumerate(atom_features):
            num = len(atoms)
            num_atoms_per_graph.append(num)
            atom_graph_indices += [i] * num

        atom_graph_indices = np.array(atom_graph_indices)

        num_bonds_per_graph = []
        bond_graph_indices = []
        for i, bonds in enumerate(bond_features):
            num = len(bonds)
            num_bonds_per_graph.append(num)
            bond_graph_indices += [i] * num

        bond_graph_indices = np.array(bond_graph_indices)
    
        # Increment is accumulative number of atom of each graph, it is used to reindex
        # indices of atom in global graph, so it should be adding to pair indices apart
        # from the first graph. The first subgraph keep its atom indices in global graph.
        # In order to add increment to pair indices, each accumulative number in increment
        # should be repeat num_bonds times so that every indice in pair_indices
        # accumulative number for first graph is zeros so that should pad num_bonds of zero to increment.
        increment = np.cumsum(num_atoms_per_graph[:-1])
        increment = np.pad(
            np.repeat(increment, num_bonds_per_graph[1:]), [(num_bonds_per_graph[0], 0)])
        
        pair_indices_per_graph = np.concatenate(pair_indices, axis=0)
        pair_indices = pair_indices_per_graph + np.expand_dims(increment, axis=-1)
        atom_features = np.concatenate(atom_features, axis=0)
        bond_features = np.concatenate(bond_features, axis=0)
        state_attrs = np.concatenate(state_attrs, axis=0)
        
        # Local spherical theta phi used for EdgeNetworks, the same as NodeNetworks.      
        local_env = np.concatenate(local_env, axis=0)
    
        return (atom_features, bond_features, local_env, state_attrs, pair_indices, atom_graph_indices,
                bond_graph_indices, pair_indices_per_graph)


class  GraphBatchGeneratorFromGraphs(GraphBatchGeneratorSequence):     
    def __init__(self, graphs: List, labels: List, task_type, batch_size=32):
        self.graphs = graphs
        self.labels = labels
        self.task_type = task_type
        self.batch_size = batch_size
        self.data_size = len(labels)


    def __getitem__(self, index: int) -> tuple:
        structure_batch = self.graphs[index * self.batch_size : (index + 1) * self.batch_size]
        y_batch = np.array(self.labels[index * self.batch_size : (index + 1) * self.batch_size])

        graph_batch = self._inputs_from_graphs(structure_batch)
        x_batch = self._merge_batch(graph_batch)
        return x_batch, (y_batch)


    def _inputs_from_graphs(self, graphs_list: List) -> Set:
        """
        Parameters
        ----------
        structure_list : List
            a list of pymatgen.core.Structure
            In order to keep with Semantic space of atom2vector, MPRester should be used
            to get structures from materialsproject.

        Returns
        -------
        Set
            DESCRIPTION.

        """
        # Initialize graphs
        atom_features_list = []
        bond_features_list = []
        state_attrs_list = []
        pair_indices_list = []
        local_env_list = []
        for s in graphs_list:
            atom_features, bond_features, state_attrs, pair_indices, local_env = s
            atom_features_list.append(atom_features)
            bond_features_list.append(bond_features)
            state_attrs_list.append(state_attrs)
            pair_indices_list.append(pair_indices)
            local_env_list.append(local_env)
  
        return (
            atom_features_list,
            bond_features_list,
            local_env_list,
            state_attrs_list,
            pair_indices_list,
            )


class  GraphBatchGeneratorBase:
    def __init__(self, X_tensor, y_label, task_type, batch_size=32, is_shuffle=False):
        """
        Parameters
        ----------
        X_tensor : TYPE
            DESCRIPTION.
        y_label : TYPE
            DESCRIPTION.
        batch_size : TYPE, optional
            DESCRIPTION. The default is 64.
        is_shuffle : TYPE, optional
            DESCRIPTION. The default is False.

        Returns
        -------
        None.

        """
        self.batch_size = batch_size
        self.task_type = task_type
        self.dataset = tf.data.Dataset.from_tensor_slices((X_tensor, (y_label)))
        if is_shuffle:
            self.dataset = self.dataset.shuffle(1024)
    
    
    def generate_dataset(self):
        """
        Returns
        -------
        TYPE
            partition datas into batches, and using merge_batch track graphs to a global graph.
        """
        return self.dataset.batch(self.batch_size).map(self._merge_batch, -1)


    def _merge_batch(self, x_batch, y_batch):
        """
        Merging a batch of graphs into a disconnected graph should reindex atoms only
        features of graphs desn't be changed only merge them to one dimension of globl graph.
        reindex indices in pair_indices by adding increment of number of atoms in the batch
        atom marked with structure indice also need to be tell in globl graph.
        Parameters
        ----------
        x_batch : TYPE
            DESCRIPTION.
        y_batch : TYPE
            DESCRIPTION.

        Returns
        -------
        atom_features : TYPE
            DESCRIPTION.
        bond_features : TYPE
            DESCRIPTION.
        state_attributes : TYPE
            DESCRIPTION.
        pair_indices : TYPE
            DESCRIPTION.
        atom_partition_indices: TYPE
            DESCRIPTION.
        bond_partition_indices: TYPE
            DESCRIPTION.
        y_batch : TYPE
            DESCRIPTION.

        """
        atom_features, bond_features, local_env, state_attrs, pair_indices = x_batch
    
        # Obtain number of atoms and bonds for each graph
        num_atoms_per_graph = atom_features.row_lengths()
        num_bonds_per_graph = bond_features.row_lengths()
        # max_num_atoms = tf.reduce_max(num_atoms_per_graph)
        
        # get adjacent matrix for each graph
        # adj_matrixes = self.adjacent_matrix_batch(max_num_atoms, pair_indices)
        
        # allocate graph (structure) indice for atom and bond in global graph
        graph_indices = tf.range(len(num_atoms_per_graph))
        atom_graph_indices = tf.repeat(graph_indices, num_atoms_per_graph)
        bond_graph_indices = tf.repeat(graph_indices, num_bonds_per_graph)
    
        # Increment is accumulative number of atom of each graph, it is used to reindex
        # indices of atom in global graph, so it should be adding to pair indices apart
        # from the first graph. The first subgraph keep its atom indices in global graph.
        # In order to add increment to pair indices, each accumulative number in increment
        # should be repeat num_bonds times so that every indice in pair_indices
        # accumulative number for first graph is zeros so that should pad num_bonds of zero to increment.
        increment = tf.cumsum(num_atoms_per_graph[:-1])
        increment = tf.pad(
            tf.repeat(increment, num_bonds_per_graph[1:]), [(num_bonds_per_graph[0], 0)])
        
        pair_indices_per_graph = pair_indices.merge_dims(outer_axis=0, inner_axis=1).to_tensor()
        pair_indices = pair_indices_per_graph + tf.expand_dims(increment, axis=-1)
        atom_features = atom_features.merge_dims(outer_axis=0, inner_axis=1)
        bond_features = bond_features.merge_dims(outer_axis=0, inner_axis=1).to_tensor()
        state_attrs = state_attrs.merge_dims(outer_axis=0, inner_axis=1)
        
        # Local spherical theta phi used for EdgeNetworks, the same as NodeNetworks.
        # num_edges_per_graph = local_env.row_lengths()
        # edge_graph_indices = tf.repeat(graph_indices, num_edges_per_graph)
        # increment_edges = tf.cumsum(num_bonds_per_graph[:-1])
        # increment_edges = tf.pad(
        #     tf.repeat(increment_edges, num_edges_per_graph[1:]), [(num_edges_per_graph[0], 0)])
        
        local_env = local_env.merge_dims(outer_axis=0, inner_axis=1).to_tensor()
    
        return (atom_features, bond_features, local_env, state_attrs, pair_indices, atom_graph_indices,
                bond_graph_indices, pair_indices_per_graph), y_batch


class  GraphBatchGenerator(GraphBatchGeneratorBase):

    def _merge_batch(self, x_batch, y_batch):
        """
        Merging a batch of graphs into a disconnected graph should reindex atoms only
        features of graphs desn't be changed only merge them to one dimension of globl graph.
        reindex indices in pair_indices by adding increment of number of atoms in the batch
        atom marked with structure indice also need to be tell in globl graph.
        Parameters
        ----------
        x_batch : TYPE
            DESCRIPTION.
        y_batch : TYPE
            DESCRIPTION.

        Returns
        -------
        atom_features : TYPE
            DESCRIPTION.
        bond_features : TYPE
            DESCRIPTION.
        state_attributes : TYPE
            DESCRIPTION.
        pair_indices : TYPE
            DESCRIPTION.
        atom_partition_indices: TYPE
            DESCRIPTION.
        bond_partition_indices: TYPE
            DESCRIPTION.
        y_batch : TYPE
            DESCRIPTION.

        """
        atom_features, bond_features, local_env, state_attrs, pair_indices = x_batch
    
        # Obtain number of atoms and bonds for each graph
        num_atoms_per_graph = []
        for i in atom_features:
            num_atoms_per_graph.append(len(i))

        num_bonds_per_graph = []
        for i in bond_features:
            num_bonds_per_graph.append(len(i))
        # max_num_atoms = tf.reduce_max(num_atoms_per_graph)
        
        # allocate graph (structure) indice for atom and bond in global graph
        graph_indices = np.arange(len(num_atoms_per_graph))
        atom_graph_indices = np.repeat(graph_indices, num_atoms_per_graph)
        bond_graph_indices = np.repeat(graph_indices, num_bonds_per_graph)
    
        # Increment is accumulative number of atom of each graph, it is used to reindex
        # indices of atom in global graph, so it should be adding to pair indices apart
        # from the first graph. The first subgraph keep its atom indices in global graph.
        # In order to add increment to pair indices, each accumulative number in increment
        # should be repeat num_bonds times so that every indice in pair_indices
        # accumulative number for first graph is zeros so that should pad num_bonds of zero to increment.
        increment = np.cumsum(num_atoms_per_graph[:-1])
        increment = np.pad(
            np.repeat(increment, num_bonds_per_graph[1:]), [(num_bonds_per_graph[0], 0)])
        
        pair_indices_per_graph = np.concatenate(pair_indices, axis=0)
        pair_indices = pair_indices_per_graph + np.expand_dims(increment, axis=-1)
        atom_features = np.concatenate(atom_features, axis=0)
        bond_features = np.concatenate(bond_features, axis=0)
        state_attrs = np.concatenate(state_attrs, axis=0)

        # Local spherical theta phi used for EdgeNetworks, the same as NodeNetworks.        
        local_env = np.concatenate(local_env, axis=0)
    
        return (atom_features, bond_features, local_env, state_attrs, pair_indices, atom_graph_indices,
                bond_graph_indices, pair_indices_per_graph), y_batch



# -*- coding: utf-8 -*-
"""
Created on Tue Oct 12 20:19:27 2021

@author: hzx
"""


from tensorflow.keras.models import Model
from tensorflow.keras.regularizers import l2
from tensorflow.keras import layers
from EdgeNetworkLayer import EdgeMessagePassing
from PartitionPaddingLayer import PartitionPadding
from MPNNTransformer import MpnnTransformerEncoder
from Readout import Set2Set


def GraphModel(
        bond_dim,
        atom_dim=16,
        num_atom=118,
        state_dim=16,
        sp_dim=230,
        units=32,
        edge_steps=1,
        message_steps=1,
        transform_steps=1,
        num_attention_heads=8,
        dense_units=32,
        output_dim=32,
        readout_units=128,
        dropout=0.0,
        reg0=0.0,
        reg1=0.0,
        reg2=0.0,
        reg3=0.0,
        reg_rec=0.0,
        batch_size=16,
        spherical_harmonics=False,
        regression=False,
        multiclassification=None,
        ):
        atom_features = layers.Input((), dtype="int32", name="atom_features_input")
        atom_features_ = layers.Embedding(num_atom, atom_dim, dtype="float32", name="atom_features")(atom_features)
        bond_features = layers.Input((bond_dim), dtype="float32", name="bond_features")
        local_env = layers.Input((6), dtype="float32", name="local_env")
        state_attrs = layers.Input((), dtype="int32", name="state_attrs_input")   
        state_attrs_ = layers.Embedding(sp_dim, state_dim, dtype="float32", name="state_attrs")(state_attrs)

        pair_indices = layers.Input((2), dtype="int32", name="pair_indices")

        atom_graph_indices = layers.Input(
            (), dtype="int32", name="atom_graph_indices"
        )

        bond_graph_indices = layers.Input(
            (), dtype="int32", name="bond_graph_indices"
        )

        pair_indices_per_graph = layers.Input((2), dtype="int32", name="pair_indices_per_graph")

        edge_features = EdgeMessagePassing(units,
                                        edge_steps,
                                        kernel_regularizer=l2(reg0),
                                        sph=spherical_harmonics
                                        )([bond_features, local_env, pair_indices])

        x_nodes_, x_edges_, x_state = MpnnTransformerEncoder()([atom_features_, edge_features, state_attrs_, pair_indices, atom_graph_indices, bond_graph_indices])
      
        x_nodes_b = PartitionPadding(batch_size)([x_nodes_, atom_graph_indices])
        x_edges_b = PartitionPadding(batch_size)([x_edges_, bond_graph_indices])

        x_nodes = layers.BatchNormalization()(x_nodes_b)
        x_edges = layers.BatchNormalization()(x_edges_b)

        x_node = Set2Set(output_dim, kernel_regularizer=l2(reg2), recurrent_regularizer=l2(reg_rec))(x_nodes)
        x_edge = Set2Set(output_dim, kernel_regularizer=l2(reg2), recurrent_regularizer=l2(reg_rec))(x_edges)
        
        x = layers.Concatenate(axis=-1, name='concat')([x_node, x_edge, x_state])
            
        x = layers.Dense(readout_units, activation="relu", kernel_regularizer=l2(reg3), name='readout0')(x)

        if dropout:
            x = layers.Dropout(dropout, name='dropout0')(x)

        x = layers.Dense(readout_units//2, activation="relu", kernel_regularizer=l2(reg3), name='readout1')(x)

        if dropout:
            x = layers.Dropout(dropout, name='dropout1')(x)

        x = layers.Dense(readout_units//4, activation="relu", kernel_regularizer=l2(reg3), name='readout2')(x)

        if dropout:
            x = layers.Dropout(dropout, name='dropout')(x)

        if regression:
            x = layers.Dense(1, name='final')(x)
        elif multiclassification is not None:
            x = layers.Dense(multiclassification, activation="softmax", name='final_softmax')(x)
        else:
            x = layers.Dense(1, activation="sigmoid", name='final')(x)

        model = Model(
            inputs=[atom_features, bond_features, local_env, state_attrs, pair_indices, atom_graph_indices,
                    bond_graph_indices, pair_indices_per_graph],
            outputs=[x],
        )
        return model


# -*- coding: utf-8 -*-
"""
Created on Tue Dec 14 14:12:08 2021

@author: huzon
"""

from tensorflow.keras.utils import to_categorical
from .crystalgraph import LabelledCrystalGraph, GraphBatchGeneratorSequence


class GraphGenerator:
    def __init__(self, dataset, data_size=None, batch_size=16, cutoff=3.0):
        self.dataset = dataset
        self.data_size = data_size
        self.batch_size = batch_size
        self.cutoff = cutoff
        self.multiclassification = None
        self.ntarget = 1

        self.train_generator, self.valid_generator, self.test_generator = self.generators()


    def generators(self):   
        structures = self.dataset.structures
        labels = self.dataset.labels
        task_type = self.dataset.task_type

        structures, labels = self.dataset.shuffle_set(structures, labels)

        if self.data_size:
            num = self.data_size
            structures_used = structures[:num]
            labels_used = labels[:num]
            labelledgraph = LabelledCrystalGraph(cutoff=self.cutoff)
            permutation = self.dataset.permute_indices(num)
        else:
            structures_used = structures
            labels_used = labels
            self.data_size = len(labels_used)
            labelledgraph = LabelledCrystalGraph(cutoff=self.cutoff)
            permutation = self.dataset.permute_indices(len(labels_used))

        print('prepare datasets, 70% for train, 20% for valid, 10% for test.')
        print('preparing train dataset...')
        x_train_, y_train = self.dataset.prepare_train_set(structures_used, labels_used, permutation)
        x_train = labelledgraph.inputs_from_strcutre_list(x_train_)

        print('preparing valid dataset...')
        x_valid_, y_valid = self.dataset.prepare_validate_set(structures_used, labels_used, permutation)
        x_valid = labelledgraph.inputs_from_strcutre_list(x_valid_)

        print('preparing test dataset...')
        x_test_, y_test = self.dataset.prepare_test_set(structures_used, labels_used, permutation)
        x_test = labelledgraph.inputs_from_strcutre_list(x_test_)

        if self.dataset.multiclassification:
            y_train = to_categorical(y_train)
            y_valid = to_categorical(y_valid)
            y_test = to_categorical(y_test)
            self.multiclassification = len(y_train[0])
        
        if self.dataset.regression:
            self.ntarget = len(y_train[0])

        train_data = GraphBatchGeneratorSequence(*x_train, y_train, task_type, batch_size=self.batch_size)
        valid_data = GraphBatchGeneratorSequence(*x_valid, y_valid, task_type, batch_size=self.batch_size)
        test_data = GraphBatchGeneratorSequence(*x_test, y_test, task_type, batch_size=self.batch_size)

        return train_data, valid_data, test_data


    def generator(self):   
        structures = self.dataset.structures
        labels = self.dataset.labels
        task_type = self.dataset.task_type

        if labels:
            structures, labels = self.dataset.shuffle_set(structures, labels)
        else:
            structures = self.dataset.shuffle_set(structures)

        if self.data_size:
            num = self.data_size
            structures_used = structures[:num]
            if labels:
                labels_used = labels[:num]
            labelledgraph = LabelledCrystalGraph(cutoff=self.cutoff)
        else:
            structures_used = structures
            if labels:
                labels_used = labels
            self.data_size = len(labels_used)
            labelledgraph = LabelledCrystalGraph(cutoff=self.cutoff)

        print('preparing dataset...')
        x_= self.dataset.prepare_x(structures_used)
        x = labelledgraph.inputs_from_strcutre_list(x_)

        if labels:
            y = self.dataset.prepare_y(labels_used)

            if self.dataset.multiclassification:
                y = to_categorical(y)
                self.multiclassification = len(y[0])

            if self.dataset.regression:
                self.ntarget = len(y[0])

        data = GraphBatchGeneratorSequence(*x, y, task_type, batch_size=self.batch_size)

        return data



# -*- coding: utf-8 -*-
"""
Created on Mon Dec 20 15:15:31 2021

@author: huzon
"""

from .crystalgraph import LabelledCrystalGraph, GraphBatchGeneratorSequence
from .database import Dataset
from .generator import GraphGenerator


# -*- coding: utf-8 -*-
"""
Created on Wed Oct 13 14:47:13 2021

@author: hzx
"""

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import backend as K
from tensorflow.keras import layers
from .graphnetworkLayer import MessagePassing
from .edgenetworkLayer import EdgeMessagePassing
from .partitionpaddingLayer import PartitionPadding, PartitionPaddingPair
from .graphtransformer import EdgesAugmentedLayer, GraphTransformerEncoder


class MpnnMultiHeadAttention(layers.Layer):
    def __init__(self, steps,
                kernel_initializer="glorot_uniform",
                kernel_regularizer=None,
                kernel_constraint=None,
                use_bias=True,
                bias_initializer="zeros",
                bias_regularizer=None,
                bias_constraint=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.steps = steps
        self.kernel_initializer = kernel_initializer
        self.kernel_regularizer = kernel_regularizer
        self.kernel_constraint = kernel_constraint
        self.use_bias = use_bias
        self.bias_initializer = bias_initializer
        self.bias_regularizer = bias_regularizer
        self.bias_constraint = bias_constraint
        self.supports_masking = True
        
        
    def build(self, input_shape):
        self.atom_dim = input_shape[0][-1]
        self.bond_dim = input_shape[1][-1]
        self.state_dim = input_shape[2][-1]

        self.query_mpnn = MessagePassing(steps=self.steps,
                                    kernel_initializer=self.kernel_initializer,
                                    kernel_regularizer=self.kernel_regularizer,
                                    kernel_constraint=self.kernel_constraint ,
                                    bias_initializer=self.bias_initializer,
                                    bias_regularizer=self.bias_regularizer,
                                    recurrent_regularizer=None,
                                    bias_constraint=None,
                                    activation=None,
                                    name='query',
                                    )

        self.key_mpnn = MessagePassing(steps=self.steps,
                                    kernel_initializer=self.kernel_initializer,
                                    kernel_regularizer=self.kernel_regularizer,
                                    kernel_constraint=self.kernel_constraint ,
                                    bias_initializer=self.bias_initializer,
                                    bias_regularizer=self.bias_regularizer,
                                    recurrent_regularizer=None,
                                    bias_constraint=None,
                                    activation=None,
                                    name='key',
                                    )

        self.value_mpnn = MessagePassing(steps=self.steps,
                                    kernel_initializer=self.kernel_initializer,
                                    kernel_regularizer=self.kernel_regularizer,
                                    kernel_constraint=self.kernel_constraint ,
                                    bias_initializer=self.bias_initializer,
                                    bias_regularizer=self.bias_regularizer,
                                    recurrent_regularizer=None,
                                    bias_constraint=None,
                                    activation=None,
                                    name='value',
                                    )


    def call(self, inputs, mask=None):
        """
        Parameters
        ----------
        query : TYPE
            DESCRIPTION.
        key : TYPE
            DESCRIPTION.
        value : TYPE
            DESCRIPTION.
        attn_bias : TYPE, optional
            DESCRIPTION. The default is None.
        attention_mask : TYPE, optional
            DESCRIPTION. The default is None.

        Returns
        -------
        attention_output : TYPE
            DESCRIPTION.

        """  
        atom_features, bond_features, state_attrs, pair_indices, atom_graph_indices, bond_graph_indices = inputs

        # calculate Q=WqIn, K=WkIn, V=WvIn, In=Inputs
        atom_features_q, bond_features_q, state_attrs_q = self.query_mpnn([atom_features, bond_features, state_attrs, pair_indices, atom_graph_indices, bond_graph_indices])
        atom_features_k, bond_features_k, state_attrs_k = self.key_mpnn([atom_features, bond_features, state_attrs, pair_indices, atom_graph_indices, bond_graph_indices])
        atom_features_v, bond_features_v, state_attrs_v = self.value_mpnn([atom_features, bond_features, state_attrs, pair_indices, atom_graph_indices, bond_graph_indices])

        # calculate Attention(Q, K, V) = softmax((QK^T)/sqrt(d_k))V
        atom_features_k = tf.transpose(atom_features_k, perm=(0, 1))
        bond_features_k = tf.transpose(bond_features_k, perm=(0, 1))
        state_attrs_k = tf.transpose(state_attrs_k, perm=(0, 1))

        atom_features_scores = tf.matmul(atom_features_q, atom_features_k)/self.atom_dim**0.5
        atom_features_attention = tf.nn.softmax(atom_features_scores)
        atom_features_output = tf.matmul(atom_features_attention, atom_features_v)

        bond_features_scores = tf.matmul(bond_features_q, bond_features_k)/self.bond_dim**0.5
        bond_features_attention = tf.nn.softmax(bond_features_scores)
        bond_features_output = tf.matmul(bond_features_attention, bond_features_v)

        state_attrs_scores = tf.matmul(state_attrs_q, state_attrs_k)/self.state_dim**0.5
        state_attrs_attention = tf.nn.softmax(state_attrs_scores)
        state_attrs_output = tf.matmul(state_attrs_attention, state_attrs_v)

        return atom_features_output, bond_features_output, state_attrs_output
        

    def get_config(self):
        config = super().get_config()
        config.update({"num_heads": self.num_heads})
        config.update({"steps": self.steps})
        return config


class MpnnTransformerEncoder(layers.Layer):
    def __init__(self, num_heads=12,
                embed_dim=16,
                edge_embed_dim=32,
                state_embed_dim=16,
                dense_dim=32,
                kernel_initializer="glorot_uniform",
                kernel_regularizer=None,
                kernel_constraint=None,
                use_bias=True,
                bias_initializer="zeros",
                bias_regularizer=None,
                bias_constraint=None,
                activation="relu",
                 **kwargs):
        super().__init__(**kwargs)

        self.attention = MpnnMultiHeadAttention(num_heads,
                                                kernel_initializer=kernel_initializer,
                                                kernel_regularizer=kernel_regularizer,
                                                kernel_constraint=kernel_constraint,
                                                use_bias=use_bias,
                                                bias_initializer=bias_initializer,
                                                bias_regularizer=bias_regularizer,
                                                bias_constraint=bias_constraint,
                                                )
        
        self.dense_proj = keras.Sequential(
            [layers.Dense(dense_dim, activation=activation, name='act_proj'), layers.Dense(embed_dim, name='proj'),]
        )
        
        self.dense_proj_edge = keras.Sequential(
            [layers.Dense(dense_dim, activation=activation, name='act_edge_proj'), layers.Dense(edge_embed_dim, name='edge_proj'),]
        )
        
        self.dense_proj_state = keras.Sequential(
            [layers.Dense(dense_dim, activation=activation, name='act_state_proj'), layers.Dense(state_embed_dim, name='state_proj'),]
        )
        
        self.layernorm_1 = layers.LayerNormalization()
        self.layernorm_2 = layers.LayerNormalization()
        
        self.layernorm_edge_1 = layers.LayerNormalization()
        self.layernorm_edge_2 = layers.LayerNormalization()

        self.layernorm_state_1 = layers.LayerNormalization()
        self.layernorm_state_2 = layers.LayerNormalization()

        self.supports_masking = True


    def call(self, inputs, mask=None):
        atom_features_output, bond_features_output, state_attrs_output = self.attention(inputs, mask=mask)
        proj_input = self.layernorm_1(inputs[0] + atom_features_output)
        proj_input_edge = self.layernorm_edge_1(inputs[1] + bond_features_output)
        proj_input_state = self.layernorm_state_1(inputs[2] + state_attrs_output)
        return (self.layernorm_2(state_attrs_output + self.dense_proj(proj_input)),
               self.layernorm_edge_2(proj_input_edge + self.dense_proj_edge(proj_input_edge)),
               self.layernorm_state_2(proj_input_state + self.dense_proj_state(proj_input_state))
        )
    

    def get_config(self):
        config = super().get_config()
        config.update({"num_heads": self.num_heads})
        config.update({"embed_dim": self.embed_dim})
        config.update({"edge_embed_dim": self.edge_embed_dim})
        config.update({"state_embed_dim": self.state_embed_dim})
        config.update({"dense_dim": self.dense_dim})
        return config


# -*- coding: utf-8 -*-
"""
Created on Wed Oct 13 14:47:13 2021

@author: hzx
"""

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import backend as K
from tensorflow.keras import layers


class EdgesAugmentedLayer(layers.Layer):
    def __init__(self):
        super().__init__()
        self.supports_masking = True
    
    def build(self, input_shape):
        self.feature_dim = input_shape[0][-1]
        self.pair_dim = input_shape[1][-1]

    
    def compute_mask(self, inputs, mask=None):
        return None


    def call(self, inputs, mask=None):
        edge_features_batch, pair_indices_batch = inputs
        
        feature_dim = self.feature_dim
        pair_dim = self.pair_dim
        
        batch_size = tf.shape(pair_indices_batch)[0]
        max_num_atoms = tf.reduce_max(tf.reshape(pair_indices_batch, (-1, 1))) + 1
        max_num_bonds = tf.shape(pair_indices_batch)[1] 
        
        initial_matrixs = tf.zeros((batch_size, max_num_atoms, max_num_atoms, feature_dim))
        batch_index = tf.cast(tf.expand_dims(tf.repeat(tf.range(batch_size), max_num_bonds), -1), dtype=tf.int64)
        indices = tf.cast(tf.reshape(pair_indices_batch, shape=(-1, pair_dim)), dtype=tf.int64)
        indices = tf.concat([batch_index, indices], -1)
        updates = tf.reshape(edge_features_batch, shape=(-1, feature_dim))
        
        augmented_matrixs = tf.tensor_scatter_nd_add(initial_matrixs, indices, updates)
        return augmented_matrixs


class GraphMultiHeadAttention(layers.Layer):
    def __init__(self, num_heads,
                mask_value=0.,
                kernel_initializer="glorot_uniform",
                kernel_regularizer=None,
                kernel_constraint=None,
                use_bias=True,
                bias_initializer="zeros",
                bias_regularizer=None,
                bias_constraint=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.num_heads = num_heads
        self.mask_value = mask_value
        self.kernel_initializer = kernel_initializer
        self.kernel_regularizer = kernel_regularizer
        self.kernel_constraint = kernel_constraint
        self.use_bias = use_bias
        self.bias_initializer = bias_initializer
        self.bias_regularizer = bias_regularizer
        self.bias_constraint = bias_constraint
        self.supports_masking = True
        
        
    def build(self, input_shape):
        self.query_dim = input_shape[0][-1]
        self.dense_dim = input_shape[0][-1]       
        self.edge_dim = input_shape[1][-1]
        
        self.out_dim = self.num_heads * self.dense_dim
        self.edge_out_dim = self.num_heads * self.edge_dim
        
        self.query_dense = layers.Dense(
            units=self.num_heads * self.query_dim,
            use_bias=self.use_bias,
            kernel_initializer=self.kernel_initializer,
            kernel_regularizer=self.kernel_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_initializer=self.bias_initializer,
            bias_regularizer=self.bias_regularizer,
            bias_constraint=self.bias_constraint,
            name='query',
        )
        
        self.key_dense = layers.Dense(
            units=self.num_heads * self.query_dim,
            use_bias=self.use_bias,
            kernel_initializer=self.kernel_initializer,
            kernel_regularizer=self.kernel_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_initializer=self.bias_initializer,
            bias_regularizer=self.bias_regularizer,
            bias_constraint=self.bias_constraint,
            name='key',
        )
        
        self.value_dense = layers.Dense(
            units=self.num_heads * self.dense_dim,
            use_bias=self.use_bias,
            kernel_initializer=self.kernel_initializer,
            kernel_regularizer=self.kernel_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_initializer=self.bias_initializer,
            bias_regularizer=self.bias_regularizer,
            bias_constraint=self.bias_constraint,
            name='value',
        )

        self.dense_augmented = layers.Dense(
            units=self.num_heads,
            use_bias=self.use_bias,
            kernel_initializer=self.kernel_initializer,
            kernel_regularizer=self.kernel_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_initializer=self.bias_initializer,
            bias_regularizer=self.bias_regularizer,
            bias_constraint=self.bias_constraint,
            name='augmented',
        )
        
        self.dense_value = layers.Dense(
            units=self.num_heads,
            use_bias=self.use_bias,
            kernel_initializer=self.kernel_initializer,
            kernel_regularizer=self.kernel_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_initializer=self.bias_initializer,
            bias_regularizer=self.bias_regularizer,
            bias_constraint=self.bias_constraint,
            name='dense_value',
        )
        
        self.dense_output = layers.Dense(
            units=self.query_dim,
            use_bias=self.use_bias,
            kernel_initializer=self.kernel_initializer,
            kernel_regularizer=self.kernel_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_initializer=self.bias_initializer,
            bias_regularizer=self.bias_regularizer,
            bias_constraint=self.bias_constraint,
            name='output',
        )

        self.dense_edge_output = layers.Dense(
            units=self.edge_dim,
            use_bias=self.use_bias,
            kernel_initializer=self.kernel_initializer,
            kernel_regularizer=self.kernel_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_initializer=self.bias_initializer,
            bias_regularizer=self.bias_regularizer,
            bias_constraint=self.bias_constraint,
            name='edge_output',
        )

        self.activation_g = layers.Activation(activation='sigmoid')

        self.built = True


    def compute_mask(self, inputs, mask=None):
        # query mask
        query_mask = None
        scores_mask = None
        if mask is not None:
            query_mask = mask[0]
            q_mask = K.cast(query_mask, K.dtype(inputs[0]))
            q_mask = tf.expand_dims(q_mask, axis=-1)
            v_mask = K.permute_dimensions(q_mask, pattern=(0, 2, 1))
            scores_mask = tf.matmul(q_mask, v_mask)
            scores_mask = tf.expand_dims(scores_mask, axis=-1)
            scores_mask = K.any(K.not_equal(scores_mask, self.mask_value), axis=-1)

        return query_mask, scores_mask


    def call(self, inputs, mask=None):
        """
        Parameters
        ----------
        query : TYPE
            DESCRIPTION.
        key : TYPE
            DESCRIPTION.
        value : TYPE
            DESCRIPTION.
        attn_bias : TYPE, optional
            DESCRIPTION. The default is None.
        attention_mask : TYPE, optional
            DESCRIPTION. The default is None.

        Returns
        -------
        attention_output : TYPE
            DESCRIPTION.

        """  
        query, edges_augmented = inputs
        batch_size = tf.shape(query)[0]

        if mask is not None:
            if mask[0] is not None:
                boolean_mask = mask[0]

        # calculate Q=WqIn, K=WkIn, V=WvIn, In=Inputs
        Query = self.query_dense(query)
        Key = self.key_dense(query)
        Value = self.value_dense(query)

        # calculate Attention(Q, K, V) = softmax((QK^T)/sqrt(d_k))V
        Query = tf.reshape(Query, shape=(batch_size, -1, self.num_heads, self.query_dim))
        Key = tf.reshape(Key, shape=(batch_size, -1, self.num_heads, self.query_dim))
        Value = tf.reshape(Value, shape=(batch_size, -1, self.num_heads, self.dense_dim))
        
        Query = tf.transpose(Query, perm=(0, 2, 1, 3))
        Key = tf.transpose(Key, perm=(0, 2, 3, 1))
        Value = tf.transpose(Value, perm=(0, 2, 1, 3))

        scores = tf.matmul(Query, Key)/self.query_dim**0.5
        attn_bias = self.dense_augmented(edges_augmented)
        attn_bias = tf.transpose(attn_bias, perm=(0, 3, 1, 2))
        scores = scores + attn_bias
        # scores = tf.matmul(scores, attn_bias)
        
        g_weights = self.dense_value(edges_augmented)
        g_weights = tf.transpose(g_weights, perm=(0, 3, 1 , 2))
        g_weights = self.activation_g(g_weights)

        # masking scores with -inf for softmax calculation
        if mask is not None:
            if mask[0] is not None:
                boolean_mask = tf.repeat(tf.expand_dims(boolean_mask, 1), self.num_heads, 1)
                q_mask = tf.expand_dims(K.cast(boolean_mask, K.dtype(query)), -1)
                v_mask = tf.transpose(q_mask, perm=(0, 1, 3, 2))
                scores_mask = tf.matmul(q_mask, v_mask)
                        
                masked_scores = scores * scores_mask - (1 - scores_mask) * 1e12
                g_weights = g_weights * scores_mask

                V = tf.einsum('ijkl,ijlm->ijkm', g_weights, Value)
        
                attention = tf.nn.softmax(masked_scores)
                attention = attention * scores_mask
                attention_output = tf.matmul(attention, V)
        else:
            masked_scores = scores
            V = tf.einsum('ijkl,ijlm->ijkm', g_weights, Value)
            attention = tf.nn.softmax(masked_scores)
            attention_output = tf.matmul(attention, V)

        attention_output = tf.reshape(attention_output, shape=(batch_size, -1, self.out_dim))
        attention_output = self.dense_output(attention_output)

        edges_output = tf.transpose(scores, perm=(0, 3, 2, 1))
        edges_output = self.dense_edge_output(edges_output)

        return attention_output, edges_output
        

    def get_config(self):
        config = super().get_config()
        config.update({"num_heads": self.num_heads})
        config.update({"mask_value": self.mask_value})
        return config

        
class GraphTransformerEncoder(layers.Layer):
    def __init__(self, num_heads=12,
                embed_dim=10,
                edge_embed_dim=16,
                dense_dim=32,
                kernel_initializer="glorot_uniform",
                kernel_regularizer=None,
                kernel_constraint=None,
                use_bias=True,
                bias_initializer="zeros",
                bias_regularizer=None,
                bias_constraint=None,
                activation="relu",
                 **kwargs):
        super().__init__(**kwargs)

        self.num_heads = num_heads
        self.embed_dim = embed_dim
        self.edge_embed_dim = edge_embed_dim
        self.dense_dim = dense_dim

        self.attention = GraphMultiHeadAttention(num_heads,
                                                kernel_initializer=kernel_initializer,
                                                kernel_regularizer=kernel_regularizer,
                                                kernel_constraint=kernel_constraint,
                                                use_bias=use_bias,
                                                bias_initializer=bias_initializer,
                                                bias_regularizer=bias_regularizer,
                                                bias_constraint=bias_constraint,
                                                )
        
        self.dense_proj = keras.Sequential(
            [layers.Dense(dense_dim, activation=activation, name='act_proj'), layers.Dense(embed_dim, name='proj'),]
        )
        
        self.dense_proj_edge = keras.Sequential(
            [layers.Dense(dense_dim, activation=activation, name='act_edge_proj'), layers.Dense(edge_embed_dim, name='edge_proj'),]
        )
        
        self.layernorm_1 = layers.LayerNormalization()
        self.layernorm_2 = layers.LayerNormalization()
        
        self.layernorm_edge_1 = layers.LayerNormalization()
        self.layernorm_edge_2 = layers.LayerNormalization()
        
        # self.layernorm_1 = layers.BatchNormalization(trainable=True)
        # self.layernorm_2 = layers.BatchNormalization(trainable=True)
        
        # self.layernorm_edge_1 = layers.BatchNormalization(trainable=True)
        # self.layernorm_edge_2 = layers.BatchNormalization(trainable=True)

        self.supports_masking = True


    def call(self, inputs, mask=None):
        attention_output, edges_output = self.attention([inputs[0], inputs[1]], mask=mask)
        proj_input = self.layernorm_1(inputs[0] + attention_output)
        proj_input_edge = self.layernorm_edge_1(inputs[1] + edges_output)
        return self.layernorm_2(proj_input + self.dense_proj(proj_input)),\
               self.layernorm_edge_2(proj_input_edge + self.dense_proj_edge(proj_input_edge))
    

    def get_config(self):
        config = super().get_config()
        config.update({"num_heads": self.num_heads})
        config.update({"embed_dim": self.embed_dim})
        config.update({"edge_embed_dim": self.edge_embed_dim})
        config.update({"dense_dim": self.dense_dim})
        return config


# -*- coding: utf-8 -*-
"""
Created on Mon Dec 20 15:15:31 2021

@author: huzon
"""


import numpy as np
from pathlib import Path
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.models import Model
from tensorflow.keras.callbacks import ModelCheckpoint, ReduceLROnPlateau, EarlyStopping
from crysnet.callbacks.cosineannealing import WarmUpCosineDecayScheduler
import matplotlib.pyplot as plt
from scipy import interp
from sklearn.metrics import roc_curve
from sklearn.metrics import auc


ModulePath = Path(__file__).parent.absolute()


class GNN:
    def __init__(self,
        model: Model,
        atom_dim=16,
        bond_dim=32,
        num_atom=118,
        state_dim=16,
        sp_dim=230,
        batch_size=16,
        regression=True,
        ntarget=1,
        multiclassification=None,
        optimizer='Adam',
        **kwargs,
        ):
        self.model = model
        self.atom_dim = atom_dim
        self.bond_dim = bond_dim
        self.num_atom = num_atom
        self.state_dim = state_dim
        self.sp_dim = sp_dim
        self.batch_size = batch_size
        self.regression = regression
        self.ntarget = ntarget
        self.multiclassification = multiclassification
        self.optimizer = optimizer

        self.gnn = model(atom_dim=atom_dim,
        bond_dim=bond_dim,
        num_atom=num_atom,
        state_dim=state_dim,
        sp_dim=sp_dim,
        batch_size=batch_size,
        regression=regression,
        multiclassification=multiclassification,
        **kwargs)


    def __getattr__(self, attr):
        return getattr(self.gnn, attr)


    def train(self, train_data, valid_data=None, test_data=None, epochs=200, lr=1e-3, warm_up=True, warmrestart=None, load_weights=False, verbose=1, checkpoints=None, save_weights_only=True, workdir=None):
        
        gnn = self.gnn
        if self.regression:
            gnn.compile(
                loss=keras.losses.MeanAbsoluteError(),
                optimizer=self.optimizer,
                metrics=[keras.metrics.MeanAbsoluteError(name='mae')],
            )
        elif self.multiclassification:
            gnn.compile(
                loss=tf.keras.losses.CategoricalCrossentropy(),
                optimizer=self.optimizer,
                metrics=[tf.keras.metrics.AUC(name="AUC")],
            )
        else:
            gnn.compile(
                loss=tf.keras.losses.BinaryCrossentropy(),
                optimizer=self.optimizer,
                metrics=[tf.keras.metrics.AUC(name="AUC")],
            )

        print(gnn.summary())
        keras.utils.plot_model(gnn, Path(workdir/"gnn_arch.png"), show_dtype=True, show_shapes=True)

        if load_weights:
            print('load weights')
            path = train_data.task_type + ".hdf5"
            if load_weights == 'default':
                best_checkpoint = Path(ModulePath/"model"/path)
            elif load_weights == 'custom':
                best_checkpoint = Path(workdir/"model"/path)
            else:
                raise ValueError('load_weights should be "default" or "custom"')
            gnn.load_weights(best_checkpoint)
        print(train_data.task_type)
        Path(workdir/"model").mkdir(exist_ok=True)
        Path(workdir/"model"/train_data.task_type).mkdir(exist_ok=True)
        if checkpoints is None:
            if self.regression:
                filepath = Path(workdir/"model"/train_data.task_type/"gnn_{epoch:02d}-{val_mae:.3f}.hdf5")
                checkpoint = ModelCheckpoint(filepath, monitor='val_mae', save_best_only=True, save_weights_only=save_weights_only, verbose=verbose, mode='min')
            else:
                filepath = Path(workdir/"model"/train_data.task_type/"gnn_{epoch:02d}-{val_AUC:.3f}.hdf5")
                checkpoint = ModelCheckpoint(filepath, monitor='val_AUC', save_best_only=True, save_weights_only=save_weights_only, verbose=verbose, mode='max')

            earlystop = EarlyStopping(monitor='val_loss', patience=200, verbose=verbose, mode='min')

            if warm_up:
                sample_count = train_data.data_size
                warmup_epoch = 5
                train_per_epoch = sample_count / self.batch_size
                warmup_steps = warmup_epoch * train_per_epoch
                restart_epoches = warmrestart

                warm_up_lr = WarmUpCosineDecayScheduler(epochs=epochs,
                                                        restart_epoches=restart_epoches,
                                                        train_per_epoch=train_per_epoch,
                                                        learning_rate_base=lr,
                                                        warmup_learning_rate=2e-6,
                                                        warmup_steps=warmup_steps,
                                                        hold_base_rate_steps=5,
                                                        )

                checkpoints = [checkpoint, warm_up_lr, earlystop]
            else:
                reduce_lr = ReduceLROnPlateau(monitor='val_loss', factor=0.2, patience=50, verbose=1, min_lr=1e-6, mode='min')
                checkpoints = [checkpoint, reduce_lr, earlystop]


        if valid_data:
            steps_per_train = int(np.ceil(train_data.data_size / self.batch_size))
            steps_per_val = int(np.ceil(valid_data.data_size / self.batch_size))
        else:
            steps_per_train = None
            steps_per_val = None

        print('gnn fit')
        history = gnn.fit(
            train_data,
            validation_data=valid_data,
            steps_per_epoch=steps_per_train,
            validation_steps=steps_per_val,
            epochs=epochs,
            verbose=verbose,
            callbacks=checkpoints,
            )

        Path(workdir/"results").mkdir(exist_ok=True)
        if self.regression:
            plot_train_regression(history, train_data.task_type, workdir)
            if test_data:
                plot_mae(gnn, test_data, workdir, name='test')
        else:
            plot_train(history, train_data.task_type, workdir)
            if test_data:
                if self.multiclassification:
                    plot_auc_multiclassification(gnn, test_data, self.multiclassification, workdir, name='test')
                else:
                    plot_auc(gnn, test_data, workdir, name='test')
        if warm_up:
            total_steps = int(epochs * sample_count / self.batch_size)
            plot_warm_up_lr(warm_up_lr, total_steps, lr, workdir)

        return gnn


    def predict_datas(self, test_data, workdir=None, load_weights='default'):
        print('load weights and predict...')
        save_file = test_data.task_type + ".hdf5"
        if load_weights:
            best_checkpoint = Path(ModulePath/"model"/save_file)
        else:
            best_checkpoint = Path(workdir/"model"/save_file)
        gnn = self.gnn()
        gnn.load_weights(best_checkpoint)
        Path(workdir/"results").mkdir(exist_ok=True)
        if self.regression:
            plot_mae(gnn, test_data, name='test')
        else:
            if self.multiclassification:
                plot_auc_multiclassification(gnn, test_data, self.multiclassification, workdir, name='test')
            else:
                plot_auc(gnn, test_data, name='test')      


    def predict(self, data, workdir=None, load_weights='default'):
        print('load weights and predict...')
        save_file = data.task_type + ".hdf5"
        if load_weights:
            best_checkpoint = Path(ModulePath/"model"/save_file)
        else:
            best_checkpoint = Path(workdir/"model"/save_file)
        gnn = self.gnn()
        gnn.load_weights(best_checkpoint)
        y_pred_keras = gnn.predict(data).ravel()
        return y_pred_keras


def plot_train(history, name, path):
    print('plot curve of training')
    plt.figure(figsize=(10, 12))
    plt.subplot(2,1,1)
    plt.plot(history.history["AUC"], label="train AUC")
    plt.plot(history.history["val_AUC"], label="valid AUC")
    plt.xlabel("Epochs", fontsize=16)
    plt.ylabel("AUC", fontsize=16)
    plt.legend(fontsize=16)
    plt.subplot(2,1,2)
    plt.plot(history.history["loss"], label="train loss")
    plt.plot(history.history["val_loss"], label="valid loss")
    plt.xlabel("Epochs", fontsize=16)
    plt.ylabel("loss", fontsize=16)
    plt.legend(fontsize=16)
    save_path = name + "_train.png"
    plt.savefig(path/"results"/save_path)


def plot_train_regression(history, name, path):
    print('plot curve of training')
    plt.figure(figsize=(10, 12))
    plt.subplot(2,1,1)
    plt.plot(history.history["mae"], label="train mae")
    plt.plot(history.history["val_mae"], label="valid mae")
    plt.xlabel("Epochs", fontsize=16)
    plt.ylabel("mae", fontsize=16)
    plt.legend(fontsize=16)
    plt.subplot(2,1,2)
    plt.plot(history.history["loss"], label="train loss")
    plt.plot(history.history["val_loss"], label="valid loss")
    plt.xlabel("Epochs", fontsize=16)
    plt.ylabel("loss", fontsize=16)
    plt.legend(fontsize=16)
    save_path = name + "_train.png"
    plt.savefig(path/"results"/save_path)


def plot_auc(gnn, test_data, path, name='test'):
    print('predict')
    name = test_data.task_type + '_' + name
    y_pred_keras = gnn.predict(test_data).ravel()
    fpr_keras, tpr_keras, _ = roc_curve(test_data.labels, y_pred_keras)
    auc_keras = auc(fpr_keras, tpr_keras)
    plt.figure(figsize=(10, 6))
    plt.plot([0, 1], [0, 1], 'k--')
    plt.plot(fpr_keras, tpr_keras, label='Keras (area = {:.3f})'.format(auc_keras))
    plt.xlabel('False positive rate')
    plt.ylabel('True positive rate')
    plt.title('ROC curve test')
    plt.legend(loc='best')
    save_path = name + "_predict" + ".png"
    plt.savefig(Path(path/"results"/save_path))


def plot_auc_multiclassification(gnn, test_data, n_classes, path, name='test'):
    print('predict')
    name = test_data.task_type + '_' + name
    y_pred_keras = gnn.predict(test_data)

    fpr = dict()
    tpr = dict()
    roc_auc = dict()
    for i in range(n_classes):
        fpr[i], tpr[i], _ = roc_curve(test_data.labels[:, i], y_pred_keras[:, i])
        roc_auc[i] = auc(fpr[i], tpr[i])

    # Compute micro-average ROC curve and ROC area
    fpr["micro"], tpr["micro"], _ = roc_curve(np.array(test_data.labels)[:, i], y_pred_keras[:, i])
    roc_auc["micro"] = auc(fpr["micro"], tpr["micro"])

    # Compute macro-average ROC curve and ROC area

    # First aggregate all false positive rates
    all_fpr = np.unique(np.concatenate([fpr[i] for i in range(n_classes)]))

    # Then interpolate all ROC curves at this points
    mean_tpr = np.zeros_like(all_fpr)
    for i in range(n_classes):
        mean_tpr += interp(all_fpr, fpr[i], tpr[i])

    # Finally average it and compute AUC
    mean_tpr /= n_classes

    fpr["macro"] = all_fpr
    tpr["macro"] = mean_tpr
    roc_auc["macro"] = auc(fpr["macro"], tpr["macro"])

    # Plot all ROC curves
    plt.figure(figsize=(10, 6))
    plt.plot(fpr["micro"], tpr["micro"],
            label='micro-average ROC curve (area = {0:0.2f})'
                ''.format(roc_auc["micro"]),
            color='deeppink', linestyle=':', linewidth=4)

    plt.plot(fpr["macro"], tpr["macro"],
            label='macro-average ROC curve (area = {0:0.2f})'
                ''.format(roc_auc["macro"]),
            color='navy', linestyle=':', linewidth=4)

    for i in range(n_classes):
        plt.plot(fpr[i], tpr[i], lw=2,
                label='ROC curve of class {0} (area = {1:0.2f})'
                ''.format(i, roc_auc[i]))

    plt.plot([0, 1], [0, 1], 'k--', lw=2)
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Some extension of Receiver operating characteristic to multi-class')
    plt.legend(loc="lower right")

    save_path = name + "_predict" + ".png"
    plt.savefig(Path(path/"results"/save_path))


def plot_mae(gnn, test_data, path, name='test'):
    print('predict')
    name = test_data.task_type + '_' + name
    y_pred_keras = gnn.predict(test_data).ravel()
    plt.figure(figsize=(10, 6))
    plt.scatter(test_data.labels, y_pred_keras)
    plt.plot([0, 8], [0, 8], 'k--')
    plt.xlim(-4, 4)
    plt.ylim(-4, 4)
    plt.xlabel("experimetal", fontsize=16)
    plt.ylabel("pred", fontsize=16)
    plt.title('predicted')
    save_path = name + "_predict" + ".png"
    plt.savefig(Path(path/"results"/save_path))


def plot_warm_up_lr(warm_up_lr, total_steps, lr, path):
    plt.plot(warm_up_lr.learning_rates)
    plt.xlabel('Step', fontsize=20)
    plt.ylabel('lr', fontsize=20)
    plt.axis([0, total_steps, 0, lr*1.1])
    # plt.xticks(np.arange(0, epochs, 1))
    plt.grid()
    plt.title('Cosine decay with warmup', fontsize=20)
    plt.savefig(Path(path/"results"/"cosine_decay.png"))    



# -*- coding: utf-8 -*-
"""
Created on Wed Nov  3 12:48:15 2021

@author: hzx
"""


from typing import Tuple
import numpy as np
from pymatgen.optimization.neighbors import find_points_in_spheres


def get_nn_info(
    structure, cutoff: float = 3.2, numerical_tol: float = 1e-8
                ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Parameters
    ----------
    structure : Structure
        DESCRIPTION.
    cutoff : float, optional
        DESCRIPTION. The default is 5.0.
    numerical_tol : float, optional
        DESCRIPTION. The default is 1e-8.

    Returns
    -------
    TYPE
        np.ndarray.
    TYPE
        np.ndarray.
    TYPE
        np.ndarray.
    TYPE
        np.ndarray.

    """
    lattice_matrix = np.ascontiguousarray(np.array(structure.lattice.matrix), dtype=float)
    pbc = np.array([1, 1, 1], dtype=int)
    r = float(cutoff)
    cart_coords = np.ascontiguousarray(np.array(structure.cart_coords), dtype=float)
    center_indices, neighbor_indices, images, distances = find_points_in_spheres(
        cart_coords, cart_coords, r=r, pbc=pbc, lattice=lattice_matrix, tol=numerical_tol
    )
    center_indices = center_indices.astype(np.int32)
    neighbor_indices = neighbor_indices.astype(np.int32)
    images = images.astype(np.int32)
    distances = distances.astype(np.float32)
    
    return center_indices, neighbor_indices, images, distances


# -*- coding: utf-8 -*-
"""
Created on Tue Oct 12 20:15:01 2021

@author: hzx
"""

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers
from tensorflow.keras import activations, initializers, regularizers, constraints


class Set2Set(layers.Layer):
    def __init__(self,
                 output_dim,
                 steps=1,
                 activation_lstm='tanh',
                 activation_recurrent='hard_sigmoid',
                 kernel_initializer='glorot_uniform',
                 recurrent_initializer='orthogonal',
                 bias_initializer="zeros",
                 kernel_regularizer=None,
                 recurrent_regularizer=None,
                 bias_regularizer=None,
                 use_bias=True,
                 kernel_constraint=None,
                 recurrent_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.supports_masking = True
        self.output_dim = output_dim
        self.steps = steps
        self.use_bias = use_bias
        self.activation_lstm = activations.get(activation_lstm)
        self.activation_recurrent = activations.get(activation_recurrent)
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.recurrent_initializer = initializers.get(recurrent_initializer)
        self.bias_initializer = initializers.get(bias_initializer)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.recurrent_regularizer = regularizers.get(recurrent_regularizer)
        self.kernel_constraint = constraints.get(kernel_constraint)
        self.kernel_constraint = constraints.get(kernel_constraint)
        self.recurrent_constraint = constraints.get(recurrent_constraint)
        self.bias_constraint = constraints.get(bias_constraint)
        self.supports_masking = True
        

    def build(self, input_shape):
        self.feature_dim = input_shape[-1]
        
        with tf.name_scope("set2set"):
            self.kernel_weight = self.add_weight(
                shape=(self.feature_dim, self.output_dim),
                initializer=self.kernel_initializer,
                regularizer=self.kernel_regularizer,
                constraint=self.kernel_constraint,
                name='set2set_kernel',
                )
            
            if self.use_bias:
                self.bias = self.add_weight(
                    shape=(self.output_dim),
                    initializer=self.bias_initializer,
                    regularizer=self.bias_regularizer,
                    constraint=self.bias_constraint,
                    name='set2set_bias',
                    )
            else:
                self.bias = None

        
        self.recurrent_kernel = self.add_weight(shape=(self.output_dim * 2,
                                                self.output_dim * 4),
                                                initializer=self.recurrent_initializer,
                                                regularizer=self.recurrent_regularizer,
                                                constraint=self.recurrent_constraint,
                                                name='recurrent_kernel',
                                                )
        
        self.recurrent_a = self.add_weight(shape=(self.output_dim * 1,),
                                            initializer=self.bias_initializer,
                                            regularizer=self.recurrent_regularizer,
                                            constraint=self.recurrent_constraint,
                                            name='recurrent_a',
                                            )

        self.recurrent_b = self.add_weight(shape=(self.output_dim * 1,),
                                            initializer=self.bias_initializer,
                                            regularizer=self.recurrent_regularizer,
                                            constraint=self.recurrent_constraint,
                                            name='recurrent_b',
                                            )

        self.recurrent_c = self.add_weight(shape=(self.output_dim * 2,),
                                            initializer=self.bias_initializer,
                                            regularizer=self.recurrent_regularizer,
                                            constraint=self.recurrent_constraint,
                                            name='recurrent_c',
                                            )

        self.built = True


    def call(self, inputs, mask=None, edge_mode=False, **kwargs):
        features = inputs
        batch_size = tf.shape(features)[0
                                        ]
        
        # if mask is not None:
        #     mask 
        
        if edge_mode:
            features = tf.reshape(features, (batch_size, -1, self.feature_dim))
        features = tf.matmul(features, self.kernel_weight) + self.bias
        
        # Set2Set embedding
        q_star = tf.reduce_sum(tf.zeros_like(features), axis=1, keepdims=True)  
        c = tf.zeros_like(q_star)  
        q_star = tf.concat([q_star, q_star], -1)  

        # set2set convert a list of vectors to a vector
        for i in range(self.steps):
            # q_t = LSTM(q∗_t−1)
            q, c = self._lstm(q_star, c)
            # print('q:',q.shape)
            # e_i,t = scalar(m_i,q_t)
            e = tf.einsum('ijk,ilk->ijl', features, q) 
            # print('e:',e.shape)
            # calculate a_i,t by softmax ei,t
            a = tf.nn.softmax(e, axis=1)  
            a = tf.tile(a, [1, 1, self.output_dim]) 
            # print(a.shape,features.shape)
            # calculated r_t by sum a_i,t and m_i
            r = tf.reduce_sum(tf.multiply(a, features), axis=1, keepdims=True)  
            q_star = tf.concat([q, r], -1)  

        return tf.reshape(q_star, [-1, self.output_dim * 2])  


    def _lstm(self, h, c):
        z = tf.matmul(h, self.recurrent_kernel)
        
        if self.use_bias:
           z = z + tf.concat([self.recurrent_a, self.recurrent_b, self.recurrent_c], -1)
        
        z_i = z[:, :, : self.output_dim]
        z_f = z[:, :, self.output_dim : self.output_dim * 2]
        z_o = z[:, :, self.output_dim * 2 : self.output_dim * 3]
        z_z = z[:, :, self.output_dim * 3 :]
        
        i = self.activation_recurrent(z_i)
        f = self.activation_recurrent(z_f)
        o = self.activation_recurrent(z_o)

        c_out = f * c + i * self.activation_lstm(z_z)
        h_out = o * self.activation_lstm(c_out)
        return h_out, c_out


    def get_config(self):
        config = super().get_config()
        config.update({"output_dim": self.output_dim})
        config.update({"steps": self.steps})
        return config


class TransformerEncoder(layers.Layer):
    def __init__(self, num_heads=8, embed_dim=64, dense_dim=512, **kwargs):
        super().__init__(**kwargs)

        self.attention = layers.MultiHeadAttention(num_heads, embed_dim)
        self.dense_proj = keras.Sequential(
            [layers.Dense(dense_dim, activation="relu"), layers.Dense(embed_dim),]
        )
        self.layernorm_1 = layers.LayerNormalization()
        self.layernorm_2 = layers.LayerNormalization()
        self.supports_masking = True

    def call(self, inputs, mask=None):
        attention_mask = mask[:, tf.newaxis, :] if mask is not None else None
        attention_output = self.attention(inputs, inputs, attention_mask=attention_mask)
        proj_input = self.layernorm_1(inputs + attention_output)
        return self.layernorm_2(proj_input + self.dense_proj(proj_input))
    
    


# -*- coding: utf-8 -*-
"""
Created on Thu Jun  3 14:56:51 2021

@author: hzx
"""


import os
import json
import requests
from tqdm import tqdm
import numpy as np
from pathlib import Path
from typing import List, Dict, Union
from pymatgen.ext.matproj import MPRester
from pymatgen.core.structure import Structure


np.random.seed(42)


class Dataset:
    """
    Accessing materialsproject.org by api.
    """
    def __init__(self, task_type=None, data_path=None, api_key=None):
        if task_type not in ['my_regression', 'my_classification', 'my_multiclassification', 'metal', 'band_gap', 'dos_fermi', 'formation_energy', 'formation_energy_all', 'formation_energy', 'e_above_hull', 'topology', 'topology_multi', 'GVRH', 'KVRH', 'possion_ratio', 'regression', 'classification', 'multiclassification']:
            raise ValueError('Invalid task type! Input task_type should be my_regression, my_classification, my_multiclassification, regression, classification, multiclassification, metal, band_gap, dos_fermi, formation_energy, e_above_hull, formation_energy_all, formation_energy, topology, GVRH, KVRH, possion_ratio.')
        
        self.task_type = task_type
        self.multiclassification = False
        
        if task_type == 'band_gap':
            self.dataset_file = Path(data_path/"datas"/"dataset_gap.json")
            self.regression = True
        elif task_type == 'dos_fermi':
            self.dataset_file = Path(data_path/"datas"/"dataset_dos.json")
            self.regression = True
        elif task_type == 'formation_energy':
            self.dataset_file = Path(data_path/"datas"/"dataset_fe.json")
            self.regression = True
        elif task_type == 'formation_energy_all':
            self.dataset_file = Path(data_path/"datas"/"dataset_fe_all.json")
            self.regression = True
        elif task_type == 'e_above_hull':
            self.dataset_file = Path(data_path/"datas"/"dataset_ehull.json")
            self.regression = True
        elif task_type == 'metal':
            self.dataset_file = Path(data_path/"datas"/"dataset_metal.json")
            self.regression = False
        elif task_type == 'topology':
            self.dataset_file = Path(data_path/"datas"/"dataset_tp.json")
            self.regression = False
        elif task_type == 'GVRH':
            self.dataset_file = Path(data_path/"datas"/"dataset_GVRH.json")
            self.regression = True
        elif task_type == 'KVRH':
            self.dataset_file = Path(data_path/"datas"/"dataset_KVRH.json")
            self.regression = True
        elif task_type == 'possion_ratio':
            self.dataset_file = Path(data_path/"datas"/"dataset_pr.json")
            self.regression = True
        elif task_type == 'regression':
            self.dataset_file = Path(data_path/"datas"/"dataset_regression.json")
            self.regression = True
        elif task_type == 'classification':
            self.dataset_file = Path(data_path/"datas"/"dataset_classification.json")
            self.regression = False
        elif task_type == 'topology_multi':
            self.dataset_file = Path(data_path/"datas"/"dataset_tp_multi.json")
            self.regression = False
            self.multiclassification = True
        elif task_type == 'multiclassification':
            self.dataset_file = Path(data_path/"datas"/"dataset_multiclassification.json")
            self.regression = False
            self.multiclassification = True
        elif task_type == 'my_regression':
            self.dataset_file = Path(data_path/"datas"/"dataset_myr.json")
            self.regression = True
        elif task_type == 'my_classificaiton':
            self.dataset_file = Path(data_path/"datas"/"dataset_myc.json")
            self.regression = False
        elif task_type == 'my_multiclassification':
            self.dataset_file = Path(data_path/"datas"/"dataset_mym.json")
            self.regression = False
            self.multiclassification = True
        
        self.structures = None
        self.labels = None

        if self.dataset_file.exists():    
            with open(self.dataset_file, 'r') as f:
                entries = json.load(f)
            if isinstance(entries[0], tuple):
                self.structures = [Structure.from_dict(s[0]) for s in entries]
                self.labels = [s[1] for s in entries]
                self.dataset = zip(self.structures, self.labels)
                self.permuted_indices = self._permute_indices(self.labels)
                self.datasize = len(self.labels)
            else:
                self.structures = [Structure.from_dict(s) for s in entries]
                self.labels = None
                self.dataset = self.structures
                self.permuted_indices = self._permute_indices(self.structures)
                self.datasize = len(self.structures)

        if api_key is not None:
            self.mpr = MPRester(api_key)

    
    def prepare_datas(self,
                      criteria = {
                              "e_above_hull": {"$lte": 0.}
                                  },
                      properties = [
                              "structure",
                              # "final_energy",
                              "formation_energy_per_atom",
                                    ],
                      ):
        """
        Parameters
        ----------
        criteria : TYPE, optional
            criteria (str/dict) : pymatgen
            for example: {"formation_energy_per_atom": {"$exists": True}}
        properties : TYPE, optional
            DESCRIPTION. The default is ["structure","final_energy",].
            Properties to request for as a list. 
            For example, ["formula", "formation_energy_per_atom"] returns the formula and formation energy per atom.
        structures: TYPE, optional
            DESCRIPTION. The default is None. If you do not want to get structure dataset from materialsproject.org by MP API MPRester.
            You have prepared your dataset as form with List[your structures in pymatgen.core.Structure]
            to define the semantic space and get atom embedding vector.
        Returns
        -------
        None.

        """
        entries = self.mpr.query(criteria=criteria, properties=properties, mp_decode=True)
        structures = [s['structure'].as_dict() for s in entries]
        labels = [s[properties[1]] for s in entries]
        
        return structures, labels


    def save_datasets(self, structures: List, labels: Union[List, None]=None):
        """
        Saving structrues and label to json file to avoid re-download entries from materialsproject.org.
        """
        if isinstance(structures[0], Structure):
            structures_dict=[]
            for s in tqdm(structures, desc="Converting Structure to json dict"):
                structures_dict.append(s.as_dict())
        else:
            structures_dict = structures
        dataset = []
        if labels:
            for structure_label in zip(structures_dict, labels):
                dataset.append(structure_label)
        else:
            for structure in structures_dict:
                dataset.append(structure)            
        with open(self.dataset_file,'w') as f:
            json.dump(dataset, f)
            

    def _permute_indices(self, x):
        permuted_indices = np.random.permutation(np.arange(len(x)))
        return permuted_indices
    
    
    def permute_indices(self, x: int):
        if isinstance(x, int):
            permuted_indices = np.random.permutation(np.arange(x))
        return permuted_indices


    def shuffle_set(self, structures: List=None, labels: List=None) -> List:
        permuted_indices = np.random.permutation(np.arange(len(labels)))

        structures_shuffle = []
        labels_shuffle = []

        if labels is None:
            for index in permuted_indices:
                structures_shuffle.append(structures[index])
            return structures_shuffle
        for index in permuted_indices:
                structures_shuffle.append(structures[index])
                labels_shuffle.append(labels[index])
        return structures_shuffle, labels_shuffle


    def prepare_train_set(self, structures: List, labels: List=None, permutation=None) -> List:
        if structures:
            self.structures = structures
        if labels:
            self.labels = labels
        if permutation is not None:
           self.permuted_indices = permutation 

        train_index = self.permuted_indices[: int(len(self.permuted_indices) * 0.7)]
        x_train = []
        y_train = []
        for index in train_index:
            x_train.append(self.structures[index])
            y_train.append(self.labels[index])
        return x_train, y_train


    def prepare_validate_set(self, structures: List, labels: List=None, permutation=None) -> List:
        if structures:
            self.structures = structures
        if labels:
            self.labels = labels
        if permutation is not None:
           self.permuted_indices = permutation 

        valid_index = self.permuted_indices[int(len(self.permuted_indices) * 0.7) : int(len(self.permuted_indices) * 0.9)]
        x_valid = []
        y_valid = []
        for index in valid_index:
            x_valid.append(self.structures[index])
            y_valid.append(self.labels[index])
        return x_valid, y_valid


    def prepare_test_set(self, structures: List, labels: List=None, permutation=None) -> List:
        if structures:
            self.structures = structures
        if labels:
            self.labels = labels
        if permutation is not None:
           self.permuted_indices = permutation 

        # test_index = self.permuted_indices
        test_index = self.permuted_indices[int(len(self.permuted_indices) * 0.9) :]
        x_test = []
        y_test = []
        for index in test_index:
            x_test.append(self.structures[index])
            y_test.append(self.labels[index])
        return x_test, y_test


    def prepare_x(self, structures: List) -> List:
        if structures:
            self.structures = structures
        else:
            raise ValueError('strucutres should not be None.')
        return self.structure
    
    
    def prepare_y(self, labels: List=None) -> List:
        self.labels = labels
        return self.labels


class TopologicalDataset(Dataset):
    def __init__(self, token=None, api_key=None):
        self.mpr = MPRester(api_key)
        self.hostname = 'materiae.iphy.ac.cn/'
        if token is not None:
            self.token = token
        else:
            raise ValueError('token should not be None, get your token from "materiae.iphy.ac.cn"')
        
        
    def query_Materiae(self) -> List:
        """
        Returns
        -------
        Dict
            dict of datas from Materiae.

        """
        hostname = self.hostname
        token = self.token
        url = "http://%s/api/materials" % hostname
        headers = {'Authorization': 'Bearer %s' % token}    
        response = requests.get(url=url, headers=headers)
        materials = response.json()['materials']
        return materials
    

    def topological_classfication(self):
        """
        Returns
        -------
        strurtures : List of pymatgen.core.struture.Structure
            strcutures used for graph networks.
        labels : List of topological classfications.
            classification labels for supervised graph networks.

        """
        hostname = self.hostname
        token = self.token
        url = "http://%s/api/materials" % hostname
        headers = {'Authorization': 'Bearer %s' % token}    
        response = requests.get(url=url, headers=headers)
        materials = response.json()['materials']
        
        strurtures = []
        labels = []
        for material in tqdm(materials):
            url = "http://%s/api/materials/%s" % (hostname, material['id'])
            headers = {'Authorization': 'Bearer %s' % token}
            response = requests.get(url=url, headers=headers)
            mp_id = response.json()['mp_id']
            # nsoc_topo_class = response.json()['nsoc_topo_class']
            soc_topo_class = response.json()['soc_topo_class']
            is_topo = 0
            if soc_topo_class != 'Triv_Ins' and soc_topo_class != '':
                is_topo = 1
            structure = self.mpr.get_structure_by_material_id(mp_id, 
                                                              final=True, 
                                                              conventional_unit_cell=False)
            strurtures.append(structure.as_dict())
            labels.append(is_topo)
            print(structure.formula,' soc_topo_class: ', soc_topo_class,' is_topo: ', is_topo)
        return strurtures, labels


    def topological_multiclassfication(self):
        """
        Returns
        -------
        strurtures : List of pymatgen.core.struture.Structure
            strcutures used for graph networks.
        labels : List of topological classfications.
            classification labels for supervised graph networks.

        """
        hostname = self.hostname
        token = self.token
        url = "http://%s/api/materials" % hostname
        headers = {'Authorization': 'Bearer %s' % token}    
        response = requests.get(url=url, headers=headers)
        materials = response.json()['materials']
        
        strurtures = []
        labels = []
        for material in tqdm(materials):
            url = "http://%s/api/materials/%s" % (hostname, material['id'])
            headers = {'Authorization': 'Bearer %s' % token}
            response = requests.get(url=url, headers=headers)
            mp_id = response.json()['mp_id']
            # nsoc_topo_class = response.json()['nsoc_topo_class']
            soc_topo_class = response.json()['soc_topo_class']
            topo = 0
            if soc_topo_class == 'Triv_Ins' or soc_topo_class == '':
                topo = 0
            elif soc_topo_class == 'HSL_SM':
                topo = 1
            elif soc_topo_class == 'HSP_SM':
                topo = 2
            elif soc_topo_class == 'TCI':
                topo = 3
            elif soc_topo_class == 'TI':
                topo = 4
            structure = self.mpr.get_structure_by_material_id(mp_id, 
                                                          final=True, 
                                                          conventional_unit_cell=False)
            strurtures.append(structure.as_dict())
            labels.append(topo)
            print(structure.formula,' soc_topo_class: ', soc_topo_class,' is_topo: ', topo)
        return strurtures, labels


    def get_properties_Materiae(self, p: str) -> Dict:
        """
        Get properties from Materiae dataset.
        --properties:
             'id',
             'formula',
             'elements',
             'elements_num',
             'mp_id',
             'icsd_ids',
             'nelec',
             'nsites',
             'spacegroup.number',
             'spacegroup.lattice_type',
             'nsoc_efermi',
             'nsoc_dos',
             'nsoc_fermi_dos',
             'nsoc_dos_gap',
             'nsoc_band',
             'nsoc_topo_class',
             'nsoc_sym_ind',
             'nsoc_ind_group',
             'soc_efermi',
             'soc_dos',
             'soc_fermi_dos',
             'soc_dos_gap',
             'soc_band',
             'soc_topo_class',
             'soc_sym_ind',
             'soc_ind_group'
        Returns
        -------
        None.

        """
        hostname = self.hostname
        token = self.token
        url = "http://%s/api/materials" % hostname
        headers = {'Authorization': 'Bearer %s' % token}    
        response = requests.get(url=url, headers=headers)
        materials = response.json()['materials']
        
        datas = {}
        for material in materials:
            url = "http://%s/api/materials/%s" % (self.hostname, material['id'])
            headers = {'Authorization': 'Bearer %s' % token}
            response = requests.get(url=url, headers=headers) 
            mp_id = response.json()['mp_id']
            datas[mp_id] = response.json()[p]
        
        return datas
    
        
    def get_topology_info(self, params):     
        #params = {'nsoc_topo_class': 'GM_SM'}
        params = {'fields': params}
        hostname = self.hostname
        token = self.token
        url = "http://%s/api/materials" % hostname
        headers = {'Authorization': 'Bearer %s' % token}    
        response = requests.get(url=url, params=params, headers=headers)
        materials = response.json()['materials']
        return materials
        

    def download_Materiae(self):
        hostname = self.hostname
        token = self.token
        url = "http://%s/api/materials" % hostname
        headers = {'Authorization': 'Bearer %s' % token}    
        response = requests.get(url=url, headers=headers)
        materials = response.json()['materials']
        datas = {}
        for material in materials:
            url = "http://%s/api/materials/%s" % (self.hostname, material['id'])
            headers = {'Authorization': 'Bearer %s' % token}
            response = requests.get(url=url, headers=headers) 
            
            mp_id = response.json()['mp_id']
            datas[mp_id] = response.json()



# -*- coding: utf-8 -*-
"""
Created on Tue Dec  7 13:26:59 2021

@author: huzon
source code from tensorflow_graphics
"""

import numpy as np
import tensorflow as tf
from typing import Tuple
from .tensor import TensorLike


def _double_factorial_loop_body(n, result, two):
  result = tf.where(tf.greater_equal(n, two), result * n, result)
  return n - two, result, two


def _double_factorial_loop_condition(n, result, two):
  return tf.cast(tf.math.count_nonzero(tf.greater_equal(n, two)), tf.bool)


def double_factorial(n: TensorLike) -> TensorLike:
  n = tf.convert_to_tensor(value=n)

  two = tf.ones_like(n) * 2
  result = tf.ones_like(n)
  _, result, _ = tf.while_loop(
      cond=_double_factorial_loop_condition,
      body=_double_factorial_loop_body,
      loop_vars=[n, result, two])
  return result


def factorial(n: TensorLike) -> TensorLike:
  n = tf.convert_to_tensor(value=n)

  return tf.exp(tf.math.lgamma(n + 1))


def generate_l_m_permutations(
    max_band: int,
    name: str = "spherical_harmonics_generate_l_m_permutations") -> Tuple[TensorLike, TensorLike]:
  with tf.name_scope(name):
    degree_l = []
    order_m = []
    for degree in range(0, max_band + 1):
      for order in range(-degree, degree + 1):
        degree_l.append(degree)
        order_m.append(order)
    return (tf.convert_to_tensor(value=degree_l),
            tf.convert_to_tensor(value=order_m))


def _evaluate_legendre_polynomial_pmm_eval(m, x):
  pmm = tf.pow(1.0 - tf.pow(x, 2.0), tf.cast(m, dtype=x.dtype) / 2.0)
  ones = tf.ones_like(m)
  pmm *= tf.cast(
      tf.pow(-ones, m) * double_factorial(2 * m - 1),
      dtype=pmm.dtype)
  return pmm


def _evaluate_legendre_polynomial_loop_cond(x, n, l, m, pmm, pmm1):
  return tf.cast(tf.math.count_nonzero(n <= l), tf.bool)


def _evaluate_legendre_polynomial_loop_body(x, n, l, m, pmm, pmm1):
  n_float = tf.cast(n, dtype=x.dtype)
  m_float = tf.cast(m, dtype=x.dtype)
  pmn = (x * (2.0 * n_float - 1.0) * pmm1 - (n_float + m_float - 1) * pmm) / (
      n_float - m_float)
  pmm = tf.where(tf.less_equal(n, l), pmm1, pmm)
  pmm1 = tf.where(tf.less_equal(n, l), pmn, pmm1)
  n += 1
  return x, n, l, m, pmm, pmm1


def _evaluate_legendre_polynomial_loop(x, m, l, pmm, pmm1):
  n = m + 2
  x, n, l, m, pmm, pmm1 = tf.while_loop(
      cond=_evaluate_legendre_polynomial_loop_cond,
      body=_evaluate_legendre_polynomial_loop_body,
      loop_vars=[x, n, l, m, pmm, pmm1])
  return pmm1


def _evaluate_legendre_polynomial_branch(l, m, x, pmm):
  pmm1 = x * (2.0 * tf.cast(m, dtype=x.dtype) + 1.0) * pmm
  # if, l == m + 1 return pmm1, otherwise lift to the next band.
  res = tf.where(
      tf.equal(l, m + 1), pmm1,
      _evaluate_legendre_polynomial_loop(x, m, l, pmm, pmm1))
  return res


def evaluate_legendre_polynomial(degree_l: TensorLike,
                                 order_m: TensorLike,
                                 x: TensorLike) -> TensorLike:
  degree_l = tf.convert_to_tensor(value=degree_l)
  order_m = tf.convert_to_tensor(value=order_m)
  x = tf.convert_to_tensor(value=x)

  pmm = _evaluate_legendre_polynomial_pmm_eval(order_m, x)
  return tf.where(
      tf.equal(degree_l, order_m), pmm,
      _evaluate_legendre_polynomial_branch(degree_l, order_m, x, pmm))


def _spherical_harmonics_normalization(l, m, var_type=tf.float64):
  l = tf.cast(l, dtype=var_type)
  m = tf.cast(m, dtype=var_type)
  numerator = (2.0 * l + 1.0) * factorial(l - tf.abs(m))
  denominator = 4.0 * np.pi * factorial(l + tf.abs(m))
  return tf.sqrt(numerator / denominator)


def _evaluate_spherical_harmonics_branch(degree,
                                         order,
                                         theta,
                                         phi,
                                         sign_order,
                                         var_type=tf.float64):
  sqrt_2 = tf.constant(1.41421356237, dtype=var_type)
  order_float = tf.cast(order, dtype=var_type)
  tmp = sqrt_2 * _spherical_harmonics_normalization(
      degree, order, var_type) * evaluate_legendre_polynomial(
          degree, order, tf.cos(theta))
  positive = tmp * tf.cos(order_float * phi)
  negative = tmp * tf.sin(order_float * phi)
  return tf.where(tf.greater(sign_order, 0), positive, negative)


def evaluate_spherical_harmonics(
    degree_l: TensorLike,
    order_m: TensorLike,
    theta: TensorLike,
    phi: TensorLike,
    name: str = "spherical_harmonics_evaluate_spherical_harmonics") -> TensorLike:    # pylint: disable=line-too-long

  with tf.name_scope(name):
    degree_l = tf.convert_to_tensor(value=degree_l)
    order_m = tf.convert_to_tensor(value=order_m)
    theta = tf.convert_to_tensor(value=theta)
    phi = tf.convert_to_tensor(value=phi)

    var_type = theta.dtype
    sign_m = tf.math.sign(order_m)
    order_m = tf.abs(order_m)
    zeros = tf.zeros_like(order_m)
    result_m_zero = _spherical_harmonics_normalization(
        degree_l, zeros, var_type) * evaluate_legendre_polynomial(
            degree_l, zeros, tf.cos(theta))
    result_branch = _evaluate_spherical_harmonics_branch(
        degree_l, order_m, theta, phi, sign_m, var_type)
    return tf.where(tf.equal(order_m, zeros), result_m_zero, result_branch)



# -*- coding: utf-8 -*-
"""
Created on Tue Dec  7 13:26:10 2021

@author: huzon
"""

from typing import Union, Sequence

import numpy as np
import tensorflow as tf

Integer = Union[int, np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16,
                np.uint32, np.uint64]
Float = Union[float, np.float16, np.float32, np.float64]
TensorLike = Union[Integer, Float, Sequence, np.ndarray, tf.Tensor, tf.Variable]


from setuptools import setup
from setuptools import find_packages


setup(
    name='crysnet',
    version='0.0.2',
    description='Labelled Graph Networks for machine learning of crystal.',
    long_description='using mpnn or graphtransformer to learn',
    long_description_content_type='text/markdown',
    author='Zongxiang Hu',
    author_email='huzongxiang@yahoo.com',
    download_url='https://github.com/huzongxiang/CrysNetwork',
    license='BSD',
    install_requires=['numpy', "scikit-learn"],
    packages=find_packages(),
    package_data={
        "crysnet": ["*.json", "*.md", "*.png"],
    },
    include_package_data=True,
    keywords=["materials", "science", "machine", "learning", "deep", "graph", "networks", "neural", "transformer", "massagepassing", "topology", "tight", "bingding", "twisted", "graphene"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.8",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
)




# -*- coding: utf-8 -*-
"""
Created on Wed Oct 13 14:47:13 2021

@author: hzx
"""

import tensorflow as tf
from tensorflow.keras import layers


class PartitionPadding(layers.Layer):
    def __init__(self, batch_size, **kwargs):
        super().__init__(**kwargs)
        self.batch_size = batch_size


    def call(self, inputs):
        features, graph_indices = inputs

        # Obtain subgraphs
        features = tf.dynamic_partition(
            features, graph_indices, self.batch_size
        )

        # Pad and stack subgraphs
        num_features = [tf.shape(f)[0] for f in features]
        max_num = tf.reduce_max(num_features)
        features_padded = tf.stack(
            [
                tf.pad(f, [(0, max_num - n), (0, 0)])
                for f, n in zip(features, num_features)
            ],
            axis=0,
        )
        
        # Remove empty subgraphs (usually for last batch)
        nonempty_examples = tf.where(tf.reduce_sum(features_padded, (1, 2)) != 0)
        nonempty_examples = tf.squeeze(nonempty_examples, axis=-1)

        features_batch =  tf.gather(features_padded, nonempty_examples, axis=0)

        return features_batch
    
    
    def get_config(self):
        config = super().get_config()
        config.update({"batch": self.batch_size})
        return config
    

class PartitionPaddingPair(layers.Layer):
    def __init__(self, batch_size, **kwargs):
        super().__init__(**kwargs)
        self.batch_size = batch_size


    def call(self, inputs):
        features, graph_indices = inputs

        # Obtain subgraphs
        features = tf.dynamic_partition(
            features, graph_indices, self.batch_size
        )

        # Pad and stack subgraphs
        num_features = [tf.shape(f)[0] for f in features]
        max_num = tf.reduce_max(num_features)
        features_padded = tf.stack(
            [
                tf.pad(f, [(0, max_num - n), (0, 0)])
                for f, n in zip(features, num_features)
            ],
            axis=0,
        )
       
        
        # Remove empty subgraphs (usually for last batch)
        nonempty_examples = tf.unique(graph_indices)[0]

        features_batch =  tf.gather(features_padded, nonempty_examples, axis=0)

        return features_batch
        

    def get_config(self):
        config = super().get_config()
        config.update({"batch_size": self.batch_size})
        return config